/*
	CCountedDictionary.cp
	
	A subclass of CDictionary that can find entries by entry number.
	This is intended for AppleEvent Object Model support, and might not 
	be useful on non-Mac platforms.
	
	by James Jennings
	May 5, 1999
*/

#include "CCountedDictionary.h"

unsigned long CCountedDictionary::CountEntries()
{
	// Since this is an intensive operation, we remember
	// the result from the first time we do it.
	if (mNumberOfEntries>0) return mNumberOfEntries;
	
	mData.seekg(mFirstEntry);
#if 0
	// this is the STL way: it's slow and it doesn't work
	istream_iterator<char> start(mData), stop;
	mNumberOfEntries = count(start, stop ,0x0A);
#elif 0
	// This is the brute force way: It's maybe twice as fast, but it doesn't work.
	// (returns value about half what it should)
	// Needs to be optimized.
	// Expected result: LE has 9,334, EL has 13,492
	// Actual result: LE 4533 in ~5 seconds, EL 6368 in ~5 seconds.
	do {
		mNumberOfEntries++;
		NextEntry();
	} while(CanDoNext());
#elif 0
	// Count the new lines. Simple.
	// LE result: 9334 entries counted in 3.88333 seconds.
	// EL result: 13492 entries counted in 5.65 seconds.
	while(mData.ignore(10000, '\n')) {
		mNumberOfEntries++;
	}
#else
	// What I really wanted to do: build a list.
	// (LE: 9335 entries counted in 2.48333 seconds.)
	// (EL: 13493 entries counted in 3.43333 seconds.)
	// (LE: 9335 entries counted in 2.5 seconds.)
	// (EL: 13493 entries counted in 3.45 seconds.)
	BuildList();
#endif
	
	return mNumberOfEntries;
}

void CCountedDictionary::BuildList()
{
	if (mNumberOfEntries>0) return;
	
	const unsigned long kMaximumRecordLength = 10000;
	mData.seekg(mFirstEntry);
	
	mList.reserve(15000);	// pre-allocate for efficiency
	do {
		mList.push_back(mData.tellg()-mFirstEntry);
	} while( mData.ignore(kMaximumRecordLength, '\n') );
	
	// restore the file to it's former state
	mData.clear();
	mData.seekg(mCurrent);
	
	mNumberOfEntries = mList.size();
//	mList.trim();	// is there a way to dispose of excess allocation?
}

unsigned long CCountedDictionary::GetEntryNumber()// const
{	// Find the entry number of the current entry.
	const_cast<CCountedDictionary *>(this)->BuildList();
	
	istream::off_type where = mCurrent - mFirstEntry;
	Offsets::iterator i = lower_bound(mList.begin(), mList.end(), where);
	assert(where == *i);
	
	return (i - mList.begin());
}

void CCountedDictionary::GoToEntry(unsigned long number)
{
	BuildList();
	
	if (number >= 0 && number < mList.size()) {
		istream::off_type where = mList[number];
		mCurrent = mFirstEntry + where;
	} else
		throw out_of_range("CCountedDictionary::GoToEntry index out of range");
}
