/*
	CDictionary.cp
	
	Dictionary class that encapsulates a Loglan dictionary
	and its index in a platform independent way.
	Must be written entirely in ANSI C++.
	
	by James Jennings
	Started February 15, 1999
	
*/

#include <assert.h>
#include "CDictionary.h"
#include "CParser.h"	// alternate for CEntryParser
#include <iostream>		// to declare cerr
#include <iterator>
#include <algorithm>	// to declare count()

static const unsigned char wordEnd = ',';
const unsigned char endEntry = 0x0A;
	
CDictionary::CDictionary( istream & data, istream & index )
	: mData(data), mIndex(index), mFound(true),
		 mFirstEntry(0), mLastEntry(0), mCurrent(0)
{
	// Find the istream::pos_type of the last entry.
	// We stuff some rough values into the instance variables,
	// call PreviousEntry(), and then clean up.
	mData.seekg(0, ios::end);	// go to end of the file
	mCurrent = mLastEntry = mData.tellg();
	PreviousEntry();			// back up to start of entry
	mLastEntry = mCurrent;		// remember it
	
	mCurrent = 0;				// initialize mCurrent
}
	
bool CDictionary::FindEntry( const char *inWord )
{	// indexed lookup
	mFound = false;
	
	if (CanDoFind()) {
		istream::pos_type pos(0);
	//	streampos pos(0);
		mFound = mIndex.Find(inWord, pos);
		mCurrent = pos;
	}
	return mFound;
}

void CDictionary::NextEntry()
{
	if ( !CanDoNext() ) return;
	
	mData.seekg(mCurrent);
	// read while looking for the record separator
	char ch;
	while( mData.get(ch) ) {
		if ( IsEntrySeparator(ch) ) break;
		if ( mLastEntry - mData.tellg() <= 0 ) break;
	}
	
	mCurrent = mData.tellg();
	mFound = true;
}

void CDictionary::PreviousEntry()
{
	if ( !CanDoPrevious() ) return;
	
	istream::pos_type pos = mCurrent;
	
	// back up while looking for the record separator
	char ch;
	do {
		pos -= 1L;
		if (pos - mFirstEntry <= 0) break;
		mData.seekg(pos-1L);
		ch = mData.peek();
	} while ( ! IsEntrySeparator(ch) );
	
	mCurrent = pos;
	mFound = true;
}

void CDictionary::RedirectEntry()
{
	string word;
	// if the current entry says "SEE <new word>" go to the new word
	if (GetRedirectedWord(word))
		FindEntry(word.c_str());
}

bool CDictionary::CanDoFind()
{	// true if data and index are set up
	return mData.good() && mIndex.good();
}

bool CDictionary::CanDoNext() const
{	// false if at last word
	return mData.good() && mLastEntry - mCurrent > 0;
}

bool CDictionary::CanDoPrevious() const
{	// false if at first word
	return mData.good() && mCurrent - mFirstEntry > 0;
}

bool CDictionary::CanDoRedirect() const
{
	string dummy;
	// save file position
	istream::pos_type pos = mData.tellg();
	// do it
	bool result = GetRedirectedWord(dummy);
	// restore file position
	mData.seekg(pos);
	
	return result;
}

bool CDictionary::WordFound() const
{	// false if FindWord() failed.
	return mFound;
}

istream::pos_type CDictionary::Peek() const
{	// get the current position/state
	return mCurrent;
}

void CDictionary::Seek(istream::pos_type inPeek)
{	// restore the position/state
	mCurrent = inPeek;
}

CDictionary::EIsWord CDictionary::IsWord( const char *inWord ) const
{
	// make sure the file position is current
	mData.seekg(mCurrent);
	
	return IsWord_AtFilePosition( inWord );
}

CDictionary::EIsWord CDictionary::IsWord_AtFilePosition( const char *inWord ) const
{
	EIsWord result = isWord_Yes;
	
	// save the current file position
	istream::pos_type saved = mData.tellg();
	
	// scan the word
	const char *p = inWord;
	char ch;
	while (result == isWord_Yes) {
		if (!mData.get(ch)) {
			result = isWord_No;				// hit the end of the file
			mData.clear();	// clear the eof flag
		} else if (ch==wordEnd && *p=='\0')
			break;							// exact match: it is the word
		else if (ch==wordEnd)
			result = isWord_No;				// inWord is longer than the word
		else if ((ch==' ' || ch=='-') && *p=='\0')
			// On the EL side, phrases are continued with a space.
			// On the LE side, similar entries are differentiated by a trailing hyphen.
			result = isWord_StartPhrase;	// first word of phrase
		else if (*p=='\0')
			result = isWord_No;				// hit end of inWord
		else if ( tolower(*p) != tolower(ch) ) 
			result = isWord_No;				// word didn't match
		p++;
	}
	
	// restore the current file position
	mData.seekg(saved);
	
	return result;
}

void CDictionary::GetWord( char *outWord, size_t inBufLen ) const
{
	// make sure the file position is current
	mData.seekg(mCurrent);
	
	GetWord_AtFilePosition( outWord, inBufLen );
}

void CDictionary::GetWord( string &outWord ) const
{
	// make sure the file position is current
	mData.seekg(mCurrent);
	
	GetWord_AtFilePosition( outWord );
}

void CDictionary::GetWord_AtFilePosition( char *outWord, size_t inBufLen ) const
{	// get the word at the current file position without changing the file position
	
	// save the current file position
	istream::pos_type saved = mData.tellg();
	
	// copy out the word
	char *p = outWord;
	char ch;
	while ( (p-outWord < inBufLen-1) && mData.get(ch) && (ch != wordEnd) ) {
		*(p++) = ch;
	}
	*p = '\0';
	
	// restore the current file position
	mData.seekg(saved);
}

void CDictionary::GetWord_AtFilePosition( string &outWord ) const
{	// get the word at the current file position without changing the file position
	
	// save the current file position
	istream::pos_type saved = mData.tellg();
	
	// copy out the word
	outWord.clear();
	char ch;
	while ( mData.get(ch) && (ch != wordEnd) ) {
		outWord += ch;
	}
	
	// restore the current file position
	mData.seekg(saved);
}

void CDictionary::GetDefinition( CFormatter &outText, bool looping )
{	// extract the current entry

	// save and restore mCurrent, because RedirectEntry() might change it
	istream::pos_type current = mCurrent;
	try {
		outText.BeginFormatting();
		// Report an error, if any.
		if ( !WordFound() ) {
			outText.SetStyle(style_Error);
			outText.PutString("Word not found. Next is:");
			outText.SetStyle(style_Normal);
			outText.PutNewLine();
		}
		
		// make sure the file position is current
		mData.seekg(mCurrent);
		
		if (looping) {
			// We're going to merge entries that have the same 'root'.
			// Get the root.
		//	char root[kMaxWordLength];
		//	GetWord(root, kMaxWordLength);
			string root;
			GetWord(root);
			
			// parse the data for entries that are 'close'
			while (looping) {
				GetDefinition_WithRedirect(outText);
				looping = (IsWord_AtFilePosition(root.c_str()) != isWord_No);
				if (looping) {
					// separate the definitions
					outText.PutNewLine();
					outText.PutNewLine();
				}
			}
		} else {
			GetDefinition_WithRedirect(outText);
		}
		// clean up
		outText.EndFormatting();
	}
	catch (...) {
		outText.AbortFormatting();
	}
	mCurrent = current;
}

void CDictionary::GetDefinition_WithRedirect( CFormatter &outText )
{	// Get the definition at the current file position,
	// following a redirected entry, if any,
	// and leaving the file pointer and an advanced but unredirected position.
	
	// This is an AtFilePosition function --
	// The things we do will modify mCurrent so we want to preserve it.
	istream::pos_type current = mCurrent;	
	istream::pos_type pos = mData.tellg();	
	
	try {
		mCurrent = pos;
		if (CanDoRedirect()) {
			// This is a "SEE <new word>" entry
		//	mData.seekg(pos);		// because CanDoRedirect() changes file position
			GetDefinition_AtFilePosition(outText);
			outText.PutNewLine();
			pos = mData.tellg();	// remember the end of that entry
			RedirectEntry();
			mData.seekg(mCurrent);
			GetDefinition_AtFilePosition(outText);
			mData.seekg(pos);		// leave at the end of the entry before redirecting
		} else {
			// this is a regular entry
			mData.seekg(pos);		// because CanDoRedirect() changes file position
			GetDefinition_AtFilePosition(outText);
		}
		mCurrent = current;
	} catch(...) {
		mCurrent = current;
		throw;
	}

}

void CDictionary::GetDefinition_AtFilePosition( CFormatter &outText )
{	// The real work of extracting an entry.
	// This is separate from GetDefinition() so 
	// that similar definitions can be merged.
		
	// Constants that make GetDefinition() more readable
	// "unsigned" so that 8-bit comparisons work
	const unsigned char wordEnd = ',';			// ','  (emit ',')
	const unsigned char fieldSep = '|';			// '|'  (space 1)
	const unsigned char startDefinition = 0xB2; // ''  (new line, space 3)
	const unsigned char endDefinition = 0xB3; 	// ''  (space 1)
	const unsigned char startEmphasis = 0xD6;	// ''  (set blue italic)
	const unsigned char endEmphasis = 0xC0; 	// ''  (set plain)
	const unsigned char startDefClass = 0xC7; 	// ''  (set bold)
	const unsigned char endDefClass = 0xC8; 	// ''  (set plain)
	
	// Start with the Entry's word style.
	outText.SetStyle(style_Word);
	bool flag_Word = true;	// are we processing the entry's word?
	bool flag_Translation = false; // are we processing a '*' translation?
	bool flag_Header = false;
	
	unsigned char ch; // unsigned so that 8 bit compares will work
	
	// Note: bool(mData.get()) returns true if the stream status if mData is ok.
	while( bool(mData.get((char&)ch)) && !IsEntrySeparator(ch) /*&& ch != EOF*/ ) {
	
		if (ch=='\t' /*|| ch=='-'*/) { // escape all tabs
			outText.PutChar(' ');
			
		} else if (ch==wordEnd && flag_Word) {	// end Entry's word
			outText.SetStyle(style_Normal); 
			outText.PutString(", ");
			outText.PutNewLine();
			flag_Word = false;
			flag_Header = true;	// italic text in the header is different
			
		} else if (ch=='*') {	// start translated word or phrase
			outText.SetStyle(style_Translation); 
			flag_Translation = true;
			
		} else if (ch=='-') { // continue translated phrase
			// Hyphens at the ends of words have a special meaning and are ignored here.
			if (flag_Word == true && mData.peek() == ',') continue;
			outText.PutChar(' '); // (hyphens outside of translation are converted too)
			
		} else if (ch=='_') { // escape the hyphens in translated phrases
			outText.PutChar('-');
			
		} else if (ch==fieldSep) { // field separator
			outText.SetStyle(style_Normal); // translation never spans fields.
			flag_Translation = false;
			// still need to do something here so that blank fields don't space too much
			outText.PutChar(' '); 
			
		} else if ( flag_Translation && 
					( !isalpha(ch) && ch != '-' && ch != '_') ) {
				//	(ch==' ' || ch == '/' || ch == ',' || ch == '.' || ch == '?' || ch == '!') ) {
			// We've come to the end of a '*' marked section.
			outText.SetStyle(style_Normal); 
			outText.PutChar(ch); 
			flag_Translation = false;
			
		} else if (ch=='&' && mData.peek()==',') { 
			// I don't understand what this is supposed to handle
			mData.ignore(); 
			
		} else if (ch==startDefinition) { // new line and indent
			flag_Header = false;	// header never goes past first startDefinition
			outText.PutNewLine();
			outText.PutString("   ");
			
		} else if (ch==endDefinition) { // small tab
			outText.SetStyle(style_Normal); // in case we have a translation running
			flag_Translation = false;
			outText.PutChar(' '); 
			
		} else if (ch==startEmphasis) {
			if (flag_Header)
				outText.SetStyle(style_Header); // start green italic
			else
				outText.SetStyle(style_Translation);  // start blue italic
			
		} else if (ch==endEmphasis) { // end blue italic
			outText.SetStyle(style_Normal); 
			
		} else if (ch==startDefClass) { // start bold
			outText.SetStyle(style_DefinitionName); 
			
		} else if (ch==endDefClass) { // end bold
			outText.SetStyle(style_Normal); 
			
		} else if (ch==';') { 		// extra space after ';' so "Used in" list wraps
			outText.PutChar(ch); 	// Can we target this more carefully?
			outText.PutChar(' '); 
			
		} else {	// pass on all other characters
			outText.PutChar(ch);
		}
	}
}

bool CDictionary::GetRedirectedWord( string &outWord ) const
{	// Returns true is the current entry says "SEE <new word>"
	// outWord is the new word
	char ch, see[5];
	
	mData.seekg(mCurrent);
	while (mData.get(ch) && ch != CParser::char_EndWord) ;	// find the end of the word
	
	// is it a redirected word?
	mData.get(see,5);
	bool redirect = (strncmp(see, "SEE ", 4) == 0);
	if (redirect) {
		// extract the word
		outWord.clear();
		while (  mData.get(ch) && 
				(ch != CParser::char_EndOfEntry && 
				 ch != CParser::char_EndOfEntry2)) outWord += ch;
	}
	return redirect;
}

void CDictionary::ParseEntry( CEntryParser &outParse )
{
	// make sure the file position is current
	mData.seekg(mCurrent);
	
	// give the parser our data
	outParse.Parse(mData);
}

void CDictionary::ParseEntry( CParse &outParse )
{
	// make sure the file position is current
	mData.seekg(mCurrent);
	
	// give the parser our data
	CParser theParser(mData, outParse);
	theParser.ParseEntry();
}

SDictionaryState &  CDictionary::GetState( SDictionaryState &ioState ) const
{
	ioState.current = Peek();
	ioState.wordFound = mFound;
	return ioState;
}

void CDictionary::SetState( const SDictionaryState &inState )
{
	Seek(inState.current);
	mFound = inState.wordFound;
}

