/*
	CDictionary.h
	
	Dictionary class that encapsulates a Loglan dictionary
	and its index in a platform independent way.
	Must be written entirely in ANSI C++.
	
	by James Jennings
	Started February 15, 1999
	
*/

#pragma once

#include <stdio.h>
#include <istream>
#include <string>
#include "CFormatter.h"
#include "CEntryParser.h"
#include "CParse.h"		// alternate for CEntryParser
//#include "CWord.h"
#include "CDictionaryIndex.h"

/*
	SDictionaryState:
	A "magic cookie" that remembers a dictionary's current state.
*/
typedef struct SDictionaryState {
		SDictionaryState() : current(0), wordFound(false) {}
		istream::pos_type	current;
		bool 				wordFound;
		} SDictionaryState;

/*
	CDictionary:
	Encapsulate a Loglan dictionary and its index.
	We will have two of these objects: one for LE and one for EL.
	
	File handling is done with iostreams.
*/
class CDictionary {
protected:
//	enum { kMaxWordLength = CWord::kMaxWordLength };
public:
				CDictionary( istream & data, istream & index );
	
	// public interface for the dictionary
	// This is what the platform dependant user interface calls.
	virtual bool FindEntry( const char *inWord );	// indexed lookup
	virtual void FirstEntry() { Seek( mFirstEntry ); }
	virtual void LastEntry() { Seek( mLastEntry ); }
	virtual void NextEntry();
	virtual void PreviousEntry();
	virtual void RedirectEntry();
	virtual bool CanDoFind();						// true if data and index are set up
	virtual bool CanDoNext() const;					// false if at last word
	virtual bool CanDoPrevious() const;				// false if at first word
	virtual bool CanDoRedirect() const;				// true if "SEE <word>"
	virtual bool WordFound() const;					// false if FindWord() failed.
	virtual istream::pos_type Peek() const;			// get the current position/state
	virtual void Seek(istream::pos_type inPeek);	// restore the position/state
		
	typedef enum { isWord_No, isWord_StartPhrase, isWord_Yes } EIsWord;
	virtual EIsWord IsWord( const char *inWord ) const;
	virtual void GetWord( char *outWord, size_t inBufLen /*= kMaxWordLength*/) const;
	virtual void GetWord( string &outWord ) const;
		// extract the current entry
	virtual void GetDefinition( CFormatter &outText, bool looping = true );
	
	virtual void ParseEntry( CEntryParser &outParse );
	virtual void ParseEntry( CParse &outParse );
	
	virtual SDictionaryState &	GetState( SDictionaryState &ioState ) const;
	virtual void				SetState( const SDictionaryState &inState );
	
protected:
	virtual EIsWord IsWord_AtFilePosition( const char *inWord ) const;
	virtual void GetWord_AtFilePosition( char *outWord, size_t inBufLen ) const;
	virtual void GetWord_AtFilePosition( string &outWord ) const;
	virtual void GetDefinition_WithRedirect( CFormatter &outText );
	virtual void GetDefinition_AtFilePosition( CFormatter &outText );
	virtual bool GetRedirectedWord( string &outWord ) const;
	
			bool IsEntrySeparator( char ch );
	// data file instance variables
	istream				&mData;
	istream::pos_type	mCurrent, mFirstEntry, mLastEntry;
	bool 				mFound;
	// index file instance variables
	CDictionaryIndex mIndex;
};

inline
bool CDictionary::IsEntrySeparator( char ch )
{	// Because the dictionary is a text file,
	// the record separator (new line) might change
	// depending on the platform the data files are prepared
	// on.
	//
	// Currently, the line separator is either a Unix \n or a Mac \r,
	// depending on the exact tools used to create the data file.
	//
	// I've routed all checks through this function call.
	return ch == 0x0A || ch == 0x0D;
}

