/*
	CDictionaryIndex.cp
	
	Index file for a Loglan dictionary.
	
	by James Jennings
	Started February 16, 1999 streambuf
*/

#pragma once

#include "CDictionaryIndex.h"
#include <assert.h>	// for debugging
#include <iostream>

CDictionaryIndex::CDictionaryIndex( istream &index )
	: mIndex(index), mRootSector(0), mCurrentSector(0)
{
	// find the root sector
	mIndex.seekg(0,ios::end);
	istream::off_type fileLength = mIndex.tellg() - istream::pos_type(0);
	mRootSector = fileLength / kSectorSize - 1;
	// preload the mData buffer
	LoadSector(mRootSector);
}

// === public interface ===

bool CDictionaryIndex::Find( const char *inWord, istream::pos_type &outPosition )
{	// return true if found
	mFound = true;
	
	const int kBufferLength = 256;
	char s[kBufferLength];
	CleanUpInput( inWord, s, kBufferLength );
	
	Search(reinterpret_cast<const unsigned char*>(s), mRootSector, 0);
	
	if (mFound)
		outPosition = mFoundAddress;
	else
		outPosition = mNextAddress;
	
	return mFound;
}

long CDictionaryIndex::CleanUpInput( const char *in, char *out, int bufLen )
{	// Make the input string 'canonical' so the search is more general.
	
	assert(in!=0 && out!=0);
	const char *p = in;
	char *q = out;
	// skip leading spaces
	while(*p != '\0' && isspace(*p)) p++;
	// convert to lower case
	while(*p!='\0' && q-out < bufLen-1) 
		*(q++) = tolower(*(p++));
	// trim trailing spaces
	while(q>out && isspace(*(q-1))) q--;
	// terminate the string
	*q = '\0';
	// return the new length
	assert(q-out < bufLen);
	return q-out;
}

// === search helpers ===

void	CDictionaryIndex::LoadSector(short inSector)
{
	assert(inSector >= 0 && inSector <= mRootSector);
	
	if (inSector == mCurrentSector) return;
	
	// load the requested sector into mData
	// Note that with 2-byte integers, inSector*kSectorSize will overflow without the cast.
	mIndex.seekg(static_cast<unsigned long>(inSector) * kSectorSize, ios::beg);
	// Note that mData[] is an array of unsigned chars so 8-bit compares will work,
	// but istream::read() expects an array of char.
	mIndex.read(reinterpret_cast<char*>(mData), kSectorSize);
	
	assert( mIndex.good() );
	
	mCurrentSector = inSector;
}

CDictionaryIndex::ESearchResult 
CDictionaryIndex::Search( const unsigned char *key, short sector, short level )
{
	if (level >= kMaxLevel) {
		mFound = false;
		return notFound;
	}
	
	LoadSector(sector);
	
	mNextSectorToSearch = GetSector(0); // next sector to search
	
	ESearchResult result = Look(key,1);
	if (result == keepLooking)
		result = Search( key, mNextSectorToSearch, level+1 );
	
	return result;
}

CDictionaryIndex::ESearchResult 
CDictionaryIndex::Look( const unsigned char *key, short no )
{
/*  Looks in an index node for the word. (Structure of index on request).
	Note that the results of Look() are returned by setting the member variables:
		mNextAddress: address of next word if word was not found
		mFoundAddress: address of the word, if it was found
		mNextSectorToSearch: the next sector to apply Look() to
*/
	
	short compare = static_cast<short>(*key) - GetChar(no);
	if ( compare < 0 ) {
/* This segment deals with the case where the test letter is lower in the alphabet than
	the index letter.  The word is not in this sector.  Get the next word address in case
	of ultimate failure by scanning through to the end of the word (a null character)
	and accessing the address of the index word being examined.  Return keepLooking to
	continue search*/
		while (GetChar(no)) ++no;
		mNextAddress = GetAddress(no+1);
		return keepLooking;
		
	} else if ( compare > 0 ) {
/* The test letter is higher in the alphabet, so we must search higher in the index.
	First look to see if there is a pointer in the lower byte of the word.  If not, it is
	not in this sector, and we must scan through to the end of the word to find the
	appropriate sector to search next.  We therefore check each subsequent 
	character for a pointer and follow any found until no further pointers are found
	At this point we can recover the sector to search next, and an address for 
	a word to display in case of failure. */
		short next = GetJump(no);
		if (next == 0) {	// is there a jump here?
			// There is no jump. Find end of the word in the index, following jumps.
			while (GetChar(no) != '\0') {
				if (GetJump(no) == 0) ++no;
				else no = GetJump(no);
			}
			// In this case, "mNextSectorToSearch" is 
			// the sector next higher in the alphabet
			mNextSectorToSearch = GetSector(no+1);
			mNextAddress = GetAddress(no+1);
			return keepLooking;	// redo the search
			
		} else { // if (next != 0)
			// There is a jump. Follow it and continue the search.
	/*  Otherwise there is a possibility that it is in this sector.  Update the 
		mNextSectorToSearch to the previous after the jump, and repeat the search from
		the offset address */
			mNextSectorToSearch = GetSector(next-2);
			return Look(key,next);
		}
		
	} else {	// if ( compare == 0 )
/* Otherwise there was a match. Continue the search in a similar 
	manner using the next characters in the test and index words. If
	the end of the test word has been reached, return the address
	contained in the two following words */
		if ( *key == '\0' ) {
			// We found it! Get the dictionary address.
			mFoundAddress = GetAddress(no+1);
			return found;
			
		} else {
			// Advance a letter and look again;
			++no;
			++key;
			return Look(key, no);
		}
	}
}

