/*
	CCGIFormatter.cpp
	
	A formatter for generating a web server/CGI response.
	Goals:
		 Have the template this is based on ("CGIForm.html") be easily editable.
		 Let the object be reusable.
	
	James Jennings * March 24, 2000
	
	Usage:
		// at least once:
		CCGIFormatter formatter(myOSTream);
		formatter.SetServerName("Apple Personal Web Sharing");
		formatter.SetActionPath("http://www.loglan.org/LOD/LOD.cgi");
		formatter.SetHTMLTemplate(theCGIForm.c_str());
		
		// just after a lookup:
		SetWord(theDict);
		GetDefinition(theDict, formatter);
*/

#include "CCGIFormatter.h"

const char crlf[] = "\015\012";

// Things we search the temlate for.
const char ACTION[] 	= "<!--ACTION-->";
const char WORD[] 		= "<!--WORD-->";
const char LEChecked[] 	= "<!--LEChecked-->";
const char ELChecked[] 	= "<!--ELChecked-->";
const char RESULT[] 	= "<!--RESULT-->";
const char PREVIOUS[] 	= "<!--PREVIOUS-->";
const char NEXT[] 		= "<!--NEXT-->";


CCGIFormatter::CCGIFormatter( ostream &os ) 
	:	CHTMLFormatter(os),
		whichDictionary(noDictionary),
		wordFound(false),
		serverName("WebStar 1.2"),
		actionPath("http://www.loglan.org/LOD/LOD.cgi")
{
}

	
void CCGIFormatter::SetServerName( const char * server )
{
	assert(server != 0);
	assert(strlen(server) < 32);	// general sanity check
	serverName = server;
}

void CCGIFormatter::SetActionPath( const char * path )
{
	assert(path != 0);
	assert(strlen(path) < 255);		// general sanity check
	actionPath = path;
}

void CCGIFormatter::SetHTMLTemplate( const char * temp )
{
	assert(temp != 0);
	assert(strlen(temp) < 2048);	// general sanity check
	htmlTemplate = temp;
}

void CCGIFormatter::SetWord( CDictionary & theDict, DictionaryType which )
{
	// extract current word information from the dictionary
	SDictionaryState savedState;
	theDict.GetState(savedState);
	
	whichDictionary = which;
	wordFound = theDict.WordFound();
	
	theDict.GetWord(currentWord);
	
	if (theDict.CanDoPrevious()) {
		theDict.PreviousEntry();
		theDict.GetWord(previousWord);
		theDict.SetState(savedState);
	} else {
		previousWord.clear();
	}
	
	if (theDict.CanDoNext()) {
		theDict.NextEntry();
		theDict.GetWord(nextWord);
		theDict.SetState(savedState);
	} else {
		nextWord.clear();
	}
}

/*
void CCGIFormatter::SetWord( char *curr, char *prev, char *next, 
							DictionaryType which, bool found )
{
	if (curr) currentWord = curr;
	if (prev) previousWord = prev;
	if (next) nextWord = next;
	whichDictionary = which;
	wordFound = found;
}
*/
	
void CCGIFormatter::BeginFormatting()
{
	if (wordFound)
		PutHTMLHeader(200);
	else
		PutHTMLHeader(404);
	if (htmlTemplate.size() == 0) {
		// if they forgot to supply a template, revert.
		CHTMLFormatter::BeginFormatting();
	} else {
		ProcessTemplate();
		out << header;
	}
}

void CCGIFormatter::EndFormatting()
{
	if (htmlTemplate.size() == 0) {
		CHTMLFormatter::EndFormatting();
	} else {
		out << footer;
	}
}
	
void CCGIFormatter::PutHTMLHeader( int type )
{
	switch (type) {
	case 200:
		out << "HTTP/1.0 200 OK" << crlf;
		break;
	case 404:
		out << "HTTP/1.0 404 Not Found" << crlf;
		break;
	default:
		out << "HTTP/1.0 500 Internal Server Error" << crlf;
		break;
	}
	out << "Server: " << serverName << " /ID CGI" << crlf;
	out << "MIME-Version: 1.0" << crlf;
	out << "Content-type: text/html" << crlf;
	out << crlf;
}

void CCGIFormatter::Replace( string &s, const char *oldText, const char *newText)
{
	string::size_type pos;
	
	pos = s.find(oldText);
	assert(pos != string::npos);	// assert the position is not "no position"
	
	if (pos>=0) 
		s.replace(pos, strlen(oldText), newText);
}

void CCGIFormatter::MakeLookupAnchor(string &result, 
		const char *word, DictionaryType whichDictionary)
{
	// make a lookup URL. Eg: LOD.cgi?Lookup=matma&in=LE
	if (word == 0 || *word == '\0') {
		result = "&lt;none&gt;";
	} else {
		string w = word;
		// replace any spaces with '+'
		replace(w.begin(), w.end(), ' ', '+');
		// to do: escape other characters with %xx (if necessary)
		
		result = "<A href=\"" + actionPath + "?lookup=" + w + "&in=" +
			(whichDictionary==LE ? "LE" : "EL") + "\">" + word + "</A>";
	}
}

void CCGIFormatter::ProcessTemplate()
{
	// Make a copy of the template. Fill stuff in.
	
	header = htmlTemplate;
	
	Replace(header, ACTION, actionPath.c_str());
	Replace(header, WORD, currentWord.c_str());
	
	assert(whichDictionary != noDictionary);
	if (whichDictionary == LE) {
		Replace(header, LEChecked, "CHECKED");
		Replace(header, ELChecked, "");
	} else {
		Replace(header, LEChecked, "");
		Replace(header, ELChecked, "CHECKED");
	}
	
	string temp;
	MakeLookupAnchor(temp, previousWord.c_str(), whichDictionary);
	Replace(header, PREVIOUS, temp.c_str());
	MakeLookupAnchor(temp, nextWord.c_str(), whichDictionary);
	Replace(header, NEXT, temp.c_str());
	
	// split the template into two pieces
	string::size_type pos, pos1;
	pos = header.find(RESULT);
	assert(pos != string::npos);
	pos1 = pos + strlen(RESULT);
	footer.assign( header, pos1, header.size() - pos1);
	header.erase( pos, header.size() - pos );
}

void CCGIFormatter::SendStartingPage()
{
	// Create a page without a word definition,
	// for when the user is just starting out.
	PutHTMLHeader(200);
	
	header = htmlTemplate;
	
	Replace(header, ACTION, actionPath.c_str());
	Replace(header, WORD, "");
	
	Replace(header, LEChecked, "CHECKED");
	Replace(header, ELChecked, "");
	
//	Replace(header, PREVIOUS, "...");
//	Replace(header, NEXT, "...");
	
	// send the result
	out << header;
}
