/*
	CParser.cpp
	
	A class that encapsulates all we need to know about parsing an LOD record.
	It can also be used for more sophisticated formatting.
	
	James Jennings * April 7, 2000
*/

#include "CParser.h"
#include <ctype.h>
#include <assert.h>

/* 
	StFilePos
	Level 3 parses often need to look ahead a bit.
	StFilePos saves and restores the file position.
*/
class StFilePos {
public:
	StFilePos( istream &is ) : stream(is), pos(is.tellg()) {}
	~StFilePos() { stream.seekg(pos); }
protected:
	istream &stream;
	istream::pos_type pos;
};



/*
	constructor
	inIn: an istream that's already aligned with the start of an entry
	inParse: is a subclass of CParse for recording the result.
*/
CParser::CParser( istream &inIS, CParse &inParse ) 
	: in(inIS), parse(inParse) 
{}

/*
	ParseEntry
	The top of the recursive descent.
*/
void	CParser::ParseEntry()
{
	doingHeader = false;
	doingDefinition = false;
	headerIsEmpty = true;
	redirected = false;
	
	parse.BeginParsing();
	try {
		// Rule: entry := word header [ definition ]+ [ footer ]
		
		if ( ! Word() ) throw ParseFailedException();	// word is required
		if ( ! Header() ) throw ParseFailedException();	// header is required, but may be empty
		if (redirected) return;
		if ( ! Definition() ) throw ParseFailedException();	// 1st definition is required
		while( Definition() ) {}	// remaining definitions are optional
		Footer();					// footer is optional
		
		parse.EndParsing();
	}
	catch(...) {
		parse.AbortParsing();
	}
}

/*
	Recursive Descent Parser Calls: Level 1
	Word(), Header(), Definition(), Footer()
*/
bool	CParser::Word()
{
	// the word is any text, followed by a comma
	SetPart(CParse::part_Word, CParse::part_Start);
	SetStyle(CParse::style_Word);
	
	if ( !PutUntil(char_EndWord) ) throw ParseFailedException();
	
	SetPart(CParse::part_Word, CParse::part_End);
	SetStyle(CParse::style_Normal);
	
	PutChar(char_EndWord);
	PutNewLine();
	return true;
}

bool	CParser::Header()
{
	// the header is any text, followed by a char_EndHeader
	// Not found in EL dictionary
	// Rule: header := [ [ wordInfo <unused> ] wordType attribution ]
	//					| "SEE" <loglan_word> ''
	//					| ''
	
	// check to see if the header is empty
	unsigned char ch = in.peek();
	headerIsEmpty = (ch == char_EndHeader);
	
	if (headerIsEmpty) {
		in.get();	// throw away char_EndHeader
		return true;// an empty header is considered a successful parse
	}
	{
		StFilePos pos(in);
		char buf [4];
		in.get(buf,4);
		redirected = (strncmp(buf, "SEE", 3) == 0);
	}
	
	if (redirected) {
	
		if ( ! PutUntil(' ') ) throw ParseFailedException(); // "SEE"
		PutChar(' ');
		
		SetPart(CParse::part_Redirect, CParse::part_Start);
		SetStyle(CParse::style_LoglanText);
		
		if ( ! PutUntil(char_EndOfEntry) ) throw ParseFailedException();// loglan word
		
		SetStyle(CParse::style_Normal);
		SetPart(CParse::part_Redirect, CParse::part_End);
		
	}	else {
		
		SetPart(CParse::part_Header, CParse::part_Start);
		if (WordInfo())
			if ( ! PutUntil(char_EndField) ) throw ParseFailedException(); // unused field
		if ( ! TypeField() ) 			throw ParseFailedException();
		if ( ! Attribution() ) 			throw ParseFailedException();
		SetPart(CParse::part_Header, CParse::part_End);
		
	}
	
	PutNewLine();
	
	return true;
}

bool	CParser::Definition()
{
	// the definition is any text, followed by a char_EndDefinition
	// LE Rule:	definition := [ phrase ] '' grammar '|' text '|' [ tags ] '' 
	// EL Rule:	definition := phrase grammar ' ' [ metaphor ] [ tags ] text '' 
	
	// First, we look ahead one character to see if we really have another definition
	unsigned char ch = in.peek();
	if (ch != char_StyleLoglan &&	// if there's a phrase, we'll see the style mark
		ch != char_EndPhrase) 		// if not, we'll see the delimiter
			return false;			// otherwise, we don't have a definition
	
	PutIndent();
	
	SetPart(CParse::part_Definition, CParse::part_Start);
	if ( !Phrase() ) throw ParseFailedException();
	if ( !Grammar() ) throw ParseFailedException();
	if (isLE()) {	// LE case
		if ( !DefinitionText() ) throw ParseFailedException();
		if ( !CaseTags() ) throw ParseFailedException();
	} else {		// EL case
		Metaphor();
		if ( !CaseTags() ) throw ParseFailedException();
		if ( !DefinitionText() ) throw ParseFailedException();
	}
	SetPart(CParse::part_Definition, CParse::part_End);
	PutNewLine();
	return true;
}

bool	CParser::Footer()
{
	// the footer is any text, followed by a char_EndFooter
	// (we will refine this later)
	
	// First, we look ahead one character to see if we really have another definition
	unsigned char ch = in.peek();
	if (ch != 'U')	// Look for start of "Used in:"
			return false;			// otherwise, we don't have a footer
	
	SetPart(CParse::part_Footer, CParse::part_Start);
	if ( !PutUntil(char_EndFooter) ) throw ParseFailedException();
	SetPart(CParse::part_Footer, CParse::part_End);
	PutNewLine();
	return true;
}

/*
	Recursive Descent Parser Calls: Level 2: Header
	WordInfo(), TypeField(), Attribution()
*/
bool	CParser::WordInfo()
{
	// the WordInfo is any text, followed by a char_EndField
	// Rule: wordInfo := [ affixes ] [ etymology ]
	
	// look ahead to see if we have any WordInfo
	if (in.peek() == char_EndField) { 
		in.get();	// discard the char_EndField
		return false;
	}
	
	SetPart(CParse::part_WordInfo, CParse::part_Start);
	Affixes();
	Etymology();
	SetPart(CParse::part_WordInfo, CParse::part_End);
	return true;
}


bool	CParser::TypeField()
{
	// the TypeField is any text, followed by a char_EndField
	// Rule: typeField := [ percentage ] wordType
	SetPart(CParse::part_TypeField, CParse::part_Start);
	Percentage();
	if ( !WordType() ) throw ParseFailedException();
	SetPart(CParse::part_TypeField, CParse::part_End);
	PutChar(' ');
	return true;
}


bool	CParser::Attribution()
{
	// the Attribution is any text, followed by a char_EndHeader
	// Rule: attribution := anyText '
	SetPart(CParse::part_Attribution, CParse::part_Start);
//	if ( !PutUntil(char_EndHeader) ) throw ParseFailedException();
	if ( !Author() ) throw ParseFailedException();
	if ( !Date() ) throw ParseFailedException();
	if ( !Rank() ) throw ParseFailedException();
	SetPart(CParse::part_Attribution, CParse::part_End);
	return true;
}

/*
	Recursive Descent Parser Calls: Level 2: Definition
	Phrase(), Grammar(), DefinitionText(), CaseTags()
*/
bool	CParser::Phrase()
{
	// LE Rule: phrase := [ '' loglan_text '' '' ]  |  ''
	// EL Rule: phrase := '' loglan_text ''
	unsigned char ch = in.peek();
	bool isEmpty = (ch == char_EndPhrase);	// in LE. EL is never empty
	
	SetPart(CParse::part_Phrase, CParse::part_Start);
	if ( !isEmpty ) {
		// we'll let PutChar() handle
		// char_StyleLoglan, char_EndStyleLoglan, and char_EndPhrase
		if ( ! PutUntil(char_EndStyleLoglan) ) throw ParseFailedException();
		PutChar(char_EndStyleLoglan);
	//	PutChar(' ');	// There's already a space in the data.
	}
	SetPart(CParse::part_Phrase, CParse::part_End);
	
	if (isEmpty) in.get();	// throw away the char_EndPhrase
	
	return true;
}

bool	CParser::Grammar()
{
	// The Grammar is any text which ends with a ')'.
	// LE Rule:	grammar := '(' [ [1-5] ] [a-z]+ ')' '|'
	// EL Rule:	grammar := '(' [ [1-5] ] [a-z]+ ')' ' '
	SetPart(CParse::part_Grammar, CParse::part_Start);
	if ( ! PutUntil(')') ) throw ParseFailedException();
	PutChar(')');
	SetPart(CParse::part_Grammar, CParse::part_End);
	
	// LE has a '|', EL has a ' '. Replace with a ' '.
	char ch = in.get();
	assert(ch == '|' || ch == ' ');
	PutChar(' ');
	
	return true;
}

bool	CParser::DefinitionText()
{
	// LE: The text ends with char_EndField
	// EL: The text ends with char_EndDefinition
	// There are some special formatting issues for translation key words
	
	// for now, assume LE  TO DO: handle formatting and EL
	SetPart(CParse::part_DefinitionText, CParse::part_Start);
	if ( isLE() ) {
		if ( ! PutUntil(char_EndField) ) throw ParseFailedException();
	} else {
		// The last EL definition typically ends with a char_EndOfEntry.
		// This is OK, but we must recover from it.
		try {
			if ( ! PutUntil(char_EndDefinition) ) throw ParseFailedException();
		} catch (EndOfEntryException ex) {
			in.unget();
		}
	}
	SetPart(CParse::part_DefinitionText, CParse::part_End);
	PutChar(' ');
	
	return true;
}

bool	CParser::CaseTags()
{
	// The CaseTags are any text which ends with a ']'.
	// LE Rule:	tags := '[' tag '-' tag* ']' ''
	// EL Rule:	tags := [ tag ] '[' tag '-' tag* ']' ' '
	// 			tag := [A-Z] (one of the case tag letters)
	// NOTE: We need to handle trailing spaces.
	// NOTE: Case tags are optional.
	
	bool isEmpty = ( in.peek() != '[' );	// not good enough for EL
	
	SetPart(CParse::part_CaseTags, CParse::part_Start);
	if ( !isEmpty ) {
		if ( ! PutUntil(']') ) throw ParseFailedException();
		PutChar(']');
		PutChar(' ');	// needed for EL, doesn't hurt LE
	}
	SetPart(CParse::part_CaseTags, CParse::part_End);
	
	SkipWhiteSpace();
	
	if (isLE()) {
		// LE has a ''
		unsigned char ch = in.get();
		if (ch != char_EndDefinition) throw ParseFailedException();
	}
	
	return true;
}

bool	CParser::Metaphor()
{
	// A Metaphor is a text surrounded by <...> and marked with char_StyleEmphasis
	// Rule := '' '<' english_text '>' ''
	// It is only found in the EL dictionary, and it's optional.
	
	bool isEmpty = ( in.peek() != char_StyleEmphasis );
	
	if (isEmpty) return false;
	
	SetPart(CParse::part_Metaphor, CParse::part_Start);
	SetStyle(CParse::style_Metaphor);
	
	in.get();	// pop the char_StyleEmphasis
	if ( ! PutUntil(char_EndStyleEmphasis) ) throw ParseFailedException();
	
	SetStyle(CParse::style_Normal);
	SetPart(CParse::part_Metaphor, CParse::part_End);
	
	return true;
}

/*
	Recursive Descent Parser Calls: Level 3: WordInfo()
	Affixes(), Etymology()
*/
bool	CParser::Affixes()
{
	// Affixes are lists of Loglan words that end with char_EndStyleEmphasis
	// They are optional. Sometimes we have style codes but no text in them.
	// Rule: affixes := [ '' [ affix_list ] '' ]
	bool isEmpty = true, isStyledEmpty = true;
	{
		// look ahead to see if the affix list is empty
		StFilePos pos(in);
		char c1, c2;
		in.get(c1).get(c2);
		isStyledEmpty = ((c1 == (char)char_StyleEmphasis) && 
						 (c2 == (char)char_EndStyleEmphasis));
		isEmpty = ! ((c1 == (char)char_StyleEmphasis) && 
					 (c2 != '<') && // starting an etymology?
					 (c2 != char_EndField) && // no affix and no etymology?
			 		 (c2 != (char)char_EndStyleEmphasis));
	}
	if ( isStyledEmpty ) in.ignore(2);	// purge the unwanted style marks
	if ( isEmpty ) return false;
	
	SetPart(CParse::part_Affixes, CParse::part_Start);
	SetStyle(CParse::style_Origins);
	
	if ( ! PutUntil(char_EndStyleEmphasis) ) throw ParseFailedException();
	
	SetStyle(CParse::style_Normal);
	SetPart(CParse::part_Affixes, CParse::part_End);
	PutChar(' ');
	
	return true;
}

bool	CParser::Etymology()
{
	// Etymology is any angle bracketed text followed by a char_EndField
	// Rule: etymology := '' '<' any_text '>' '|' ''   |   '|'
	// The char_EndStyleEmphasis is usually beyond the char_EndField.
	// We let the char_EndStyleEmphasis fall into the next <unused> field. 
	
	// test to see if Etymology is empty
	bool isEmpty = true;
	{
		StFilePos pos(in);
		char c1, c2;
		in.get(c1).get(c2);
		if ((c1 == (char)char_StyleEmphasis) && (c2 == '<'))
			isEmpty = false;
	}
	if (isEmpty) {
		// we must consume the field end at least
		if ( ! PutUntil(char_EndField) ) throw ParseFailedException();
		return false;
	}
	
	SetPart(CParse::part_Etymology, CParse::part_Start);
	SetStyle(CParse::style_Origins);
	
	if ( ! PutUntil(char_EndField) ) throw ParseFailedException();
	
	SetStyle(CParse::style_Normal);
	SetPart(CParse::part_Etymology, CParse::part_End);
	PutChar(' ');
	
	// ignore the remaining ugly bits
//	assert(in.peek() == (char)char_EndField);
//	in.get();
//	assert(in.peek() == (char)char_EndStyleEmphasis);
//	in.get();
	
	return true;
}

/*
	Recursive Descent Parser Calls: Level 3: TypeField()
	Percentage(), TheType()
*/
bool	CParser::Percentage()
{
	// The Percentage is one or two digits and a '%'.
	bool isEmpty = true;
	{
		StFilePos pos(in);
		char c1,c2,c3;
		in >> c1 >> c2 >> c3;
		// look for one or two digit percentages like "5%" or "43%"
		isEmpty = ! (isdigit(c1) && (c2 == '%' || (isdigit(c2) && c3 == '%') ) );
	}
	SetPart(CParse::part_Percentage, CParse::part_Start);
	if ( !isEmpty ) {
		if ( ! PutUntil('%') ) throw ParseFailedException();
		PutChar('%');
		PutChar(' ');
	}
	SetPart(CParse::part_Percentage, CParse::part_End);
	
	SkipWhiteSpace();
	if (!isEmpty) PutChar(' ');
	
	return !isEmpty;
}

bool	CParser::WordType()
{
	// WordType is any text followed by a '|'
	// Rule: wordType := anyText '|'
	SetPart(CParse::part_WordType, CParse::part_Start);
	if ( !PutUntil(char_EndField) ) throw ParseFailedException();
	SetPart(CParse::part_WordType, CParse::part_End);
	return true;
}

/*
	Recursive Descent Parser Calls: Level 3: Attribution()
	Author(), Date(), Rank()
*/
bool	CParser::Author()
{
	// Author is any text followed by a space
	SetPart(CParse::part_Author, CParse::part_Start);
	if ( !PutUntil(' ') ) throw ParseFailedException();
	SetPart(CParse::part_Author, CParse::part_End);
	PutChar(' ');
	return true;
}

bool	CParser::Date()
{
	// Date is any text followed by a space
	SetPart(CParse::part_Date, CParse::part_Start);
	if ( !PutUntil(' ') ) throw ParseFailedException();
	SetPart(CParse::part_Date, CParse::part_End);
	PutChar(' ');
	return true;
}

bool	CParser::Rank()
{
	// Rank is any text followed by an char_EndHeader
	SetPart(CParse::part_Rank, CParse::part_Start);
	if ( !PutUntil(char_EndHeader) ) throw ParseFailedException();
	SetPart(CParse::part_Rank, CParse::part_End);
	return true;
}

/*
	CParse interface
	SetPart(), PutChar(), PutUntil()
*/
void 	CParser::SetPart( CParse::EPart inPart, CParse::EPartEdge inEdge )
{
	// here we can change the styles for each part
	
	// but for now, play it straight
	parse.SetPart(inPart, inEdge);
}

void 	CParser::PutChar( char inChar )
{	
	// catch some special cases before passing it on to the CParse object
	switch(static_cast<unsigned char>(inChar)) {
	
	// there are two char_EndOfEntrys because there are two versions of the data file
	case char_EndOfEntry:
	case char_EndOfEntry2:	throw EndOfEntryException();
	case char_EndOfLine:	throw EndOfLineException();
	case char_EndField:		throw EndOfFieldException();
	
	// If we're filling a suggestion box, then we don't excape things in the definition body.
	// When we handle AppleEvent that extract definition pieces, we'll need something smarter.
//	case '-':				parse.PutChar(' ');	break;	// spaces are escaped
//	case '_':				parse.PutChar('-');	break;	// hyphens are escaped
	
	// until the database is fixed, we add spaces
//	case ';': parse.PutChar(';'); parse.PutChar(' '); break;	database has been fixed
	
	// handle some special style characters
	case char_EndPhrase: break;	// ignore: Already used in Phrase()
	case char_StyleLoglan: 
		SetStyle(CParse::style_LoglanText);
		break;
	case char_EndStyleLoglan: 
		SetStyle(CParse::style_Normal);
		break;
	case char_StyleEmphasis:
		if (doingHeader)
			SetStyle(CParse::style_Origins);
		else
			SetStyle(CParse::style_Translation);
		break;
	case char_EndStyleEmphasis:
		SetStyle(CParse::style_Normal);
		break;
	
	default:
		parse.PutChar(inChar);
		break;
	}
}

/*
	PutUntil()	utility function
	Passes characters from 'in' to 'parse' until the stopChar is reached.
	Return 'true' is the stopChar is reached, false if end of file.
	The stopChar is discarded.
*/
bool	CParser::PutUntil( char stopChar )
{
	char ch;
	while(in.get(ch)) {
		// The data file might have either Unix or Mac line separators.
		// Merge them.
		// (It might make more sense to override in.get().)
		if (ch == char_EndOfEntry2) ch = char_EndOfEntry;
		if (ch == stopChar) break;
		PutChar(ch);
	}
	return ch == stopChar;
}

/*
	SkipWhiteSpace()	utility function
	Used when we want to ignore any white space in the data file
	and enforce our own formatting.
*/
void	CParser::SkipWhiteSpace()
{
	while(in && isspace(in.peek()))
		in.get();
}
