/*
	Affix.cpp
	
	class Affix
	
	Encapsulates routines that deal with a Loglan Affix.
	
	James Jennings * April 18, 2000
*/

#pragma once

#include "Affix.h"
#include "Letter.h"
#include "CPrimSet.h"

#ifdef __MWERKS__
#pragma mark	=== static utility functions ===
#endif

bool Affix::IsCCV(const char *aff)
{
	return	Letter::IsConsonant(aff[0]) &&
			Letter::IsConsonant(aff[1]) &&
			Letter::IsVowel(aff[2]);
}

bool Affix::IsCVC(const char *aff)
{
	return	Letter::IsConsonant(aff[0]) &&
			Letter::IsVowel(aff[1]) &&
			Letter::IsConsonant(aff[2]);
}

bool Affix::IsCVV(const char *aff)
{
	return	Letter::IsConsonant(aff[0]) &&
			Letter::IsVowel(aff[1]) &&
			Letter::IsVowel(aff[2]);
}

bool Affix::IsDisyllabicCVV(const char *aff)
{	// is this a disyallabic CVV?
	if ( ! IsCVV(aff) ) return false;
	
	// test for monosyllabic cases
	if (aff[1]=='i' && aff[2]!='i') return false;	// ia, ie, io, iu
	if (aff[1]=='u' && aff[2]!='u') return false;	// ua, ue, ui, uo
	if (aff[1]!='i' && aff[2]=='i') return false;	// ai, ei, oi, ui
	if (aff[1]=='a' && aff[2]=='o') return false;	// ao
	
	return true;
}

bool Affix::IsCDoubleV(const char *aff)
{	// is this a CVV where the V == V?
	if ( ! IsCVV(aff) ) return false;
	return aff[1] == aff[2];
}

bool Affix::IsLW(const char *aff, int len )
{
	// Is this one of the standard LWs that can be combined into complexes?
	// len can be either 2 or 3, depending on whether the LW has been hyphenated.
	const char 	*lwSet[] = {
					"fur","jur","nor","nur","rar","ror","nir","ner","tor","ter",
					"for","fer","por","sor","ser","vor","ver","rer","pir","sur",
					"nun","fun", 
					0};	// zero marks the end of the array
	
	if (strlen(aff) == len)	// optimization
		for (int i=0; lwSet[i] != 0; ++i)
			if ( ! strncmp(aff, lwSet[i], len) ) return true;
	
	return false;
}

bool Affix::IsLettoral(const char *aff)
{
	// Checks if an affix is a letter word.
	// (it doesn't really catch all letter words, but just those of a particular affix type)
	// Do we want to use Word::IsLetter() instead?
	return (strstr(aff,"ai") || strstr(aff,"ei") || strstr(aff,"eo"));
	
	/*
		Real letter words are of the form:
		
		Consonants:
				Latin:		Greek:
		lower:	Cei			Ceo
		upper:	Cai			gaoCeo
		Vowels:
				Latin:		Greek:
		lower:	Vfi			Vzi
		upper:	Vma			gaoVfi
	
	*/
}

// ===========================================================================
//	 IsBorrowed		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Check for a borrowing mainly by eliminating non borrowed word forms

bool Affix::IsBorrowed(const char *aff)
{
	// Used by class ComplexMaker.
	// Note that non-standard LWs (Eg: 'vu') get treated as borrowings
	short i;
	i = strlen(aff);
	if (i==4 && !Letter::IsVowel(aff[0])) return false; // 4-letter borrowings begin with V
	if (i!=5 && i!=3) return true; // If not length 3 or 5
	
	// 5-letter primitives begin with a consonant CvCCv or CCvCv and have one at posn 4
	if (i==5 && (Letter::IsVowel(aff[0]) || Letter::IsVowel(aff[3]))) return true; 
	
	return false;
}

#ifdef __MWERKS__
#pragma mark	=== Affix methods ===
#endif
/*
Affix::Affix(const Word &w, const char *aff)
{
	Set( w, aff );
}
*/
void Affix::Set( const Word &w, bool isFinal ) 
{
	// Set the affix and form to the default for a word.
	BuildAffix(w, isFinal);
	BuildForm( w.Get().c_str(), affix.c_str() );
}

void Affix::Set( const Word &w, const char *aff, bool isFinal ) 
{
	// Set the affix to a non-default, but use the default form.
	assert( aff != 0 );
	final = isFinal;
	Word::PushUntilEndOfWord( affix, aff );
	BuildForm( w.Get().c_str(), affix.c_str() );
	type = CalculateType(affix.c_str());
}

void Affix::Set( const char *aff, const char *frm, bool isFinal ) 
{
	// Set a non-default affix and a non-default form.
	assert( aff != 0 && frm != 0 );
	final = isFinal;
	Word::PushUntilEndOfWord( affix, aff );
	form = frm;
	type = CalculateType(affix.c_str());
}

const string &	Affix::Get() const 
{
	return affix; 
}
/*
void Affix::SetForm( const char *f ) 
{
	// SetForm is not generally used unless the default form 
	// created by Set() is inappropriate.
	Word::PushUntilEndOfWord( form, f );
}
*/
const string &	Affix::GetForm() const 
{
	return form; 
}

Affix::Type	Affix::CalculateType( const char *aff )
{
	Type typ = none;
	if (IsCCV(aff)) 			typ = CCV;
	else if (IsCVC(aff)) 		typ = CVC;
	else if (IsDisyllabicCVV(aff))
		if (aff[1] == aff[2]) 	typ = diCVV;
		else 					typ = diCVv;
	else if (IsCVV(aff))
		if (aff[1] == aff[2]) 	typ = monoCVV;
		else 					typ = monoCVv;
	else if (IsLettoral(aff))	typ = letr;	// might be 3 or 6 letters
	else if (strlen(aff) == 5)	typ = CalculatePrimitiveType(aff);
	else						typ = other;
	
	assert(typ != none);
	return typ;
}

Affix::Type	Affix::CalculatePrimitiveType( const char *a )
{
	assert(a!=0 && strlen(a) == 5);
	
	if (Word::IsPredicate(a))
		if (Letter::IsConsonant(a[0]))
			if (Letter::IsConsonant(a[2]))	return CVCCV;
			else							return CCVCV;
	
	return other;
}

void	Affix::BuildAffix( const Word &w, bool isFinal )
{
	// Build the affix from the Word.
	// We set the Affix::Type at the same time.
	// Note: This code does not have to worry about most three letter affixes.
	affix = w.Get();
	type = none;
	final = isFinal;
	
	// first sort by word size:
	switch(affix.size()) {
		// one and two letter words: append an 'r' and see 
		// if it works as a three letter word
	case 1:
	case 2:	
		if (IsLW(affix.c_str(), 2)) {
			// certain special LWs get hyphenated with 'r'
			// others get 'y', below.
			affix.push_back('r');
			type = CVC;
			break;
		}
		// fall through...
	case 3:
		if (IsLettoral(affix.c_str()))
			type = letr;
		else {
			type = CVC;
			if ( ! IsLW(affix.c_str()) )
				affix.push_back('y');
		}
		break;
		
	case 5:
		type = CalculatePrimitiveType(affix.c_str());
		if ( !final )
			if (type == other) {
				// for non-primitives, we append a 'y'
				affix.push_back('y');
			} else {	// type == CCVCV or CVCCV
				// for primitives, we replace the final vowel
				affix[4] = 'y';
			}
		break;
		
	default:
		// all other non-terminal affixes end in 'y'
		if ( !final )
			affix.push_back('y');
		type = other;
		break;
	}
	
	assert(type != none);
}

void	Affix::BuildForm( const char *wd, const char *af )
{
	// Build an Affix form based on how the word and the affix are different.
	form.clear();
	
	// First check for special cases.
	// (better here than in CPrim::GetAffixForm())
	assert(af != 0);
	if (strlen(af) == 3) {
		// Forms for out-of-sequence affixes.
		static const char * cases[]={
			"fro","fra","flo",
			"mra","mro",
			"sna","sro","sku","sru",
			"cna","klu", 
			0};	// zero marks the end of the array
		// The equivalent out-of-sequence forms.
		static const char * forms[]={
			"fo/r(li)","fa/r(fu)","fo/l(ma)",
			"ma/r(ka)","mo/r(du)",
			"sa/n(ca)","so/r(di)","su/k(si)","su/r(na)",
			"ca/n(li)","ku/(t)l(a)"};
		
		// First, check for the special cases.
		for (int i = 0; cases[i] != 0; i++) {
			if (strcmp(cases[i], af) == 0) {
				form = forms[i];	// found one!
				return;
			}
		}
	}
	
	// No special cases apply. Do it straight.
	assert(wd != 0 && af != 0);
	while (*wd) {
		if (*wd == *af) {			// if the word letter matches the affix letter
			form.push_back(*wd);	// copy the letter
			wd++; af++;				// and advance
		} else {					// otherwise
			form.push_back('(');	// parenthesize what doesn't match
			while (*wd && *wd != *af) 
				form.push_back(*wd++);
			form.push_back(')');
		}
	}
	// if there is stuff left over in the affix, it's probably a hyphen
	while (*af == 'r' || *af == 'y') {
		form.push_back('+');
		form.push_back(*af++);
	}
	// We ought to be done.
	assert(*af == '\0');
}

long	Affix::GetScore() const
{
	// calculate the score based on the affix type
	assert(type >= 0 && type <= other);
	
	long score;
	const short conPlus = 20; // continuant CVC bonus (was 33 in an earlier draft)
#if 0	
	const short isFinal[11] = {132,0,154,154,176,176,0,6,   0, 6, 100};
	const short nonFinal[11] = {363,371,373,373,390,390,115,65,   115, 65, 100};
	
	if (final) score = isFinal[type];
	else score = nonFinal[type];
#else	
	// an attempt to rewrite the above so it's easier to maintain
	const short scores[][2] = { 
					// {non-final, final},
		/*	CCV		*/		{ 363,	132	},
		/*	CVC		*/		{ 371,	0	},
		/*	diCVv	*/		{ 373,	154	},
		/*	diCVV	*/		{ 373,	154	},
		/*	monoCVv	*/		{ 390,	176	},
		/*	monoCVV	*/		{ 390,	176	},
		/*	letr	*/		{ 115,	0	},
		/*	dummy	*/		{ 65,	6	},
		/*	CCVCV	*/		{ 115,	0	},//{ 0,	115	}, swapped to match makedict
		/*	CVCCV	*/		{ 65,	6	},//{ 6,	65	}, 1/10/01
		/*	other	*/		{ 100,	100	}
							};
							
	score = scores[type][final ? 1 : 0];
#endif

	if ( type==CVC && Letter::IsContinuant(affix[2]) ) score += conPlus;
	
	return score;
}

#ifdef __MWERKS__
#pragma mark	=== AffixSet methods ===
#endif

void AffixSet::Set( const Word &w, bool isFinal )
{
	// set the default affix
	set[0].Set(w, isFinal);
	count = 1;
	
	// set additional primitive affixes, if any
	const CPrim *prim = CPrimSet::Find(w.Get().c_str());
	if (prim != 0) {
	
		for (int i = 0; i < prim->AffixCount(); i++) {
		
			assert( count < kMaxCount );
			// final affixes must end in vowels
			if ( !isFinal || Letter::IsVowel(prim->Affix(i)[2]) ) {
			
			//	char buffer[CPrim::kFormLength+1];	// Ugly. Can we get rid of this?
			//	set[count].Set(prim->Affix(i), prim->GetAffixForm(i, buffer), isFinal);
				set[count].Set(w, prim->Affix(i), isFinal);
				
				count++;
			}
		}
		
	}
}

const Affix& AffixSet::Get( int i )
{
	assert(i < count);
	return set[i];
}
