/*
	Affix.h
	
	class Affix
	
	Encapsulates routines that deal with a Loglan Affix.
	
	class AffixSet
	
	Collects the possible affixes for a word.
	
	James Jennings * April 18, 2000
*/

#pragma once

#include "CPrimSet.h"
#include "Word.h"

#include <ostream>
using namespace std;

class Affix {
public:
	// Affix::Type is a list of affix types.
	// The order is important: it is used as an array index in GetScore().
	typedef enum {
			none = -1,
			CCV,
			CVC,
			diCVv,		// di-syllabic non-equal vowels
			diCVV,		// di-syllabic equal vowels
			monoCVv,	// mono-syllabic non-equal vowels
			monoCVV,	// mono-syllabic equal vowels
			letr,		// lettoral
			dummy,
			CCVCV,		// primitive
			CVCCV,		// primitive
			other
			} Type;
	
	// simple tests
	static bool		IsCVC( const char *aff );
	static bool		IsCCV( const char *aff );
	static bool		IsCVV( const char *aff );
	static bool		IsDisyllabicCVV( const char *aff );// IsCVVD
	static bool		IsCDoubleV( const char *aff ); // IsCVVO
	
	static bool		IsLW(const char *aff, int len = 3);
	static bool		IsLettoral(const char *aff);
	static bool 	IsBorrowed(const char *theString);
	
	virtual bool	IsBorrowed() const { return IsBorrowed(affix.c_str()); }
	
	// the encapsulated affix
					Affix() : type(none), final(false) {}
	
	virtual void			Set( const Word &w, bool isFinal );
	virtual void			Set( const Word &w, const char *aff, bool isFinal );
	virtual void			Set( const char *aff, const char *frm, bool isFinal );
	virtual const string & 	Get() const;
	
//	virtual void			SetForm(const char *a);
	virtual const string & 	GetForm() const;
	
	virtual Type	GetType() const { return type; }
	virtual long	GetScore() const;
	virtual bool	IsFinal() const { return final; }
	
	virtual ostream & debug(ostream &os);
	
protected:
	virtual void	BuildAffix( const Word &w, bool isFinal );
	virtual void	BuildForm( const char *w, const char *a );
	virtual Type	CalculateType( const char *aff );
	virtual Type	CalculatePrimitiveType( const char *a );
	
	Type	type;
	bool	final;
	string	affix, form;
};

inline
ostream & Affix::debug(ostream &os)
{
	os << "Affix: " << affix << ':' << form << " Score: " << GetScore() <<
		" final: " << (final ? "true" : "false") << endl;
	return os;
}

class AffixSet {
	enum { kMaxCount = 4 };
public:
						AffixSet() : count(0) {}
						
	virtual void		Set( const Word &w, bool isFinal );
	virtual const Affix& Get( int i );
	virtual int			Count() { return count; }
	
	virtual const Affix& operator[] (int i) { return Get(i); }
	
	virtual ostream & debug(ostream &os);
	
protected:
	int		count;
	Affix	set[kMaxCount];
};

inline
ostream & AffixSet::debug(ostream &os)
{
	os << "AffixSet: " << count << endl;
	for (int i=0; i<count; i++) set[i].debug(os);
	return os;
}

