/*
	CPrimSet.cpp
	
	A class for using the prim.set data.
	
	by James Jennings
*/

#include "CPrimSet.h"

static const int kBuffLen = 256;
static const int kWordLoc = 4;	// where the word is
static const int kAff1Loc = 10;	// where the 1st affix is
static const int kAff2Loc = 14;	// where the 2nd affix is
static const int kAff3Loc = 18;	// where the 3rd affix is

static const short kPrimSetLength = 1500;	// storage for CPrim objects

// We expect prim.set to be a Mac text file on any platform.
static const char kMacLineDelimiter = 0x0D;	// '\r'

// === CPrim ===

/*
	CPrim constructor
	Zero out the member data.
*/
CPrim::CPrim()
	: count(0)
{
	// zero the member data
	memset(word, 0, kPrimLength+1);
	memset(aff, 0, (kAffixLength+1)*kMaxCount);
}

/*
	GetAffixForm()
	Find the LOD style crib for the affix origin.
	(Not needed in the New Word Maker. -- JPJ)
*/
const char * CPrim::GetAffixForm( int i, char outForm[kFormLength+1] ) const
{
	assert( i>=0 && i<kMaxCount && outForm != 0 );
	
	// Forms for out of sequence affixes.
	static const char *cases[]={
		"fro","fra","flo",
		"mra","mro",
		"sna","sro","sku","sru",
		"cna","klu", 
		0};	// zero marks the end of the array
	static const char *forms[]={
		"fo/r(li)","fa/r(fu)","fo/l(ma)",
		"ma/r(ka)","mo/r(du)",
		"sa/n(ca)","so/r(di)","su/k(si)","su/r(na)",
		"ca/n(li)","ku/(t)l(a)"};
	
	// First, check for the special cases.
	for (int j = 0; cases[j] != 0; j++) {
		if (strcmp(cases[j], aff[i]) == 0) {
			assert(strlen(forms[j]) <= kFormLength);
			strcpy(outForm, forms[j]);
			return outForm;
		}
	}
	
	// It's not an out-of-sequence form. Do it straight.
	const char *wd = word, *af = aff[i];
	char *out = outForm;
	while (*wd) {
		if (*wd == *af) {		// if the word letter matches the affix letter
			*out = *wd;			// copy the letter
			out++; wd++; af++;	// and advance
		} else {				// otherwise
			*out++ = '(';		// parenthesize what doesn't match
			while (*wd && *wd != *af) *out++ = *wd++;
			*out++ = ')';
		}
	}
	*out = '\0'; // terminate the string
	assert(out - outForm <= kFormLength);
	
	return outForm;
}

/*
	global operator >> ()
	Used for reading the CPrim in from an istream.
	
	(To do: check that the istream is good before proceeding?)
*/
istream & operator >> (istream &is, CPrim &prim)
{	// read a CPrim out of an istream
	
	istream::sentry s(is);	// standard extractor error checking
	if (s) {
		// read a line of the file
		char s[kBuffLen];
		is.getline(s, kBuffLen, kMacLineDelimiter);
		long len = strlen(s);
		assert(len >= kWordLoc && len < kBuffLen-1);
		
		// copy out the word
		memmove(prim.word, &s[kWordLoc], CPrim::kPrimLength);
		
		if (s[kAff1Loc] != '-' && len > kAff1Loc) {	// if there is a 1st affix
			// copy it out
			memmove(prim.aff[0], &s[kAff1Loc], CPrim::kAffixLength); 
			prim.count = 1;
			if (s[kAff2Loc-1] == ' ' && len > kAff2Loc) {	// if there is a 2nd affix
				// copy it out
				memmove(prim.aff[1], &s[kAff2Loc], CPrim::kAffixLength);
				prim.count = 2;
				if (s[kAff3Loc-1] == ' ' && len > kAff3Loc) {	// if there is a 3nd affix
					// copy it out
					memmove(prim.aff[2], &s[kAff3Loc], CPrim::kAffixLength);
					prim.count = 3;
				}
			}
		}
	}
	
	return is;
}

// === CPrimSet ===

CPrimSet * CPrimSet::sSingleton = 0;

/*
	CPrimSet constructor
	Takes an open istream that contains the prim.set data.
*/
CPrimSet::CPrimSet(istream &is)
	: primList(0), count(0)
{
	assert(sSingleton == 0);	// we only allow one of these to be constructed
	
//	char s[kBuffLen];
	try {
		// allocate the list
		// (We might want to figure out the exact length of the list.
		//	For now, 1500 is plenty.)
		primList = new CPrim[kPrimSetLength];
		assert(primList != 0);		// should throw an exception?
		
		// rewind the istream and discard the first line
		is.seekg(0, ios::beg);
	//	is.getline(s, kBuffLen, kMacLineDelimiter);
		is.ignore(kBuffLen, kMacLineDelimiter);
		
		// read until the file is empty
		while(is) {
			// Question: How does this code fare in various \r, \n, \r\n environments?
			// We might want to specify a Mac delimiter in getline().
			
			// skip blank lines (like at the end of the file)
			if (!isprint(is.peek())) {
				is.ignore(kBuffLen, kMacLineDelimiter);
				continue;
			}
			// read a CPrim
			is >> primList[count];
			count++;
			assert(count < kPrimSetLength);
		}
	}
	catch(...) {
		// clean up if something goes wrong
		count = 0;
		if (primList)
			delete primList;
		primList = 0;
		throw;
	}
}

/*
	CPrimSet destructor
*/
CPrimSet::~CPrimSet()
{
	assert(sSingleton == this);
	delete [] primList;
}

void CPrimSet::Initialize(istream &is)
{
	assert(sSingleton == 0);
	sSingleton = new CPrimSet(is);
	assert(sSingleton != 0);
}

void CPrimSet::Dispose()
{
	if (sSingleton != 0)
		delete sSingleton;
	sSingleton = 0;
}

/*
	CPrimSet::Find()
	Find the CPrim object that corresponds to a particular word.
	Returns a nil pointer if not found.
	This is a "class function" that can be called without having access
	to an object of the class. E.g.
		CPrim * thePrim = CPrimSet::Find( "matma" );
*/
const CPrim * CPrimSet::Find( const char *word )
{
	if (Good())
		return sSingleton->FindWord(word);
	else
		return 0;
}

/*
	CPrimSet::FindWord()
	Find the CPrim object that corresponds to a particular word.
	Returns a nil pointer if not found.
	This is usually only called through the class function Find().
*/
const CPrim * CPrimSet::FindWord( const char *word ) const
{	
	// optimization: If the word is not five letters long, it's not on the list.
	if (strlen(word) != 5) return 0;
	
	int cond;
	int low, high, mid;
	low = 0;
	high = count-1;
	
	// basic binary search
	while (low <= high) {
		mid = (high+low)/2;
		cond = strcmp(word, primList[mid].Word());
		
		if (cond<0) high = mid-1;		// mid is too high
		else if (cond>0) low = mid+1;	// mid is too low
		else return & primList[mid];	// found it!
	}
	// not found: return nil
	return 0;
}
