/*
	ComplexMaker.cpp
	
	class ComplexMaker
	
	For making complex predicates.
	
	James Jennings * April 19, 2000
*/

#include "ComplexMaker.h"
#include "Letter.h"

/*
	Choose the TRACE level for debugging FindBest()
	0: Trace is off
	1: Show best score
	2: Trace each permutation
	3: Trace each affix
	4: Trace each decision.
*/
#define TRACE	0

#if TRACE > 0
	#include <iostream>
	const char *affixTypeNames[] = {
			"none",
			"CCV",
			"CVC",
			"diCVv",	// di-syllabic non-equal vowels
			"diCVV",	// di-syllabic equal vowels
			"monoCVv",	// mono-syllabic non-equal vowels
			"monoCVV",	// mono-syllabic equal vowels
			"letr",		// lettoral
			"dummy",
			"CCVCV",	// primitive
			"CVCCV",	// primitive
			"other"
			};
	static const char *TYPE( Affix::Type t ) { return affixTypeNames[t+1]; }
#endif

#ifdef __MWERKS__
#pragma mark	=== static utility functions ===
#endif

// ===========================================================================
//	 IsBadCccJoint		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// A certain number of CCCs are nonpronounceable. This checks for those.

bool ComplexMaker::IsBadCccJoint(char a, char b, char c)

/*This routine examines two words a and b to see if the last consonant of word a,
 and the first two consonants of word b are in a list of forbidden ccc's */

{	
	const char *ccc[]={	// currently 20 CCCs
		"cdz","cvl","jdj", "jtc","jts","jvr","ndj","ndz","dcm","dct",
		"dts","kdz","pdz","svl","tvl","gts", "gzb","mzb","vts","njm",
		0};	// zero marks the end of the array
	
	char tst[4];
	tst[0]=a;
	tst[1]=b;
	tst[2]=c;
	tst[3]='\0';
	
	for (int i = 0; ccc[i] != 0; ++i)
		if (strncmp(tst,ccc[i], 3) == 0) 
			return true; 
	
	return false;
}

// ===========================================================================
//	 IsValidJunction	1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// A certain number of CCs are not accurately pronounceable. This checks for those.

bool ComplexMaker::IsValidJunction(char a, char b)

/* Check if the joint between the last character of an affix a and the first character
 of the next affix b is legal*/

{	
/*the two characters cannot be identical, and the first cannot be 'h'*/

	if (a == 'h' || a == b) return false;
	
/*Both characters cannot be spirants*/

	if (Letter::IsSpirant(a) && Letter::IsSpirant(b)) return false;
	
/*If the first character is non-voiced, and the second is a voiced IsSpirant, the joint is illegal*/

	if (Letter::IsUnvoiced(a) && (b == 'j' || b == 'z')) return false;
	
/*An unvoiced consonant cannot be followed by the equivalent voiced cons.*/

	if (a == 'p' && b == 'b') return false;
	if (a == 't' && b == 'd') return false;	
	if (a == 'k' && b == 'g') return false;
	if (a == 'f' && b == 'v') return false;
	if (a == 'b' && b == 'j') return false;
	if (a == 's' && b == 'b') return false;

	return true; /*all others are valid*/
}

// ===========================================================================
//	 IsInitialPair		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// An initial pair at a junction may have to be hyphenated. Check for these

bool  ComplexMaker::IsInitialPair(char a, char b)

/*Check if the consonant pair ab is a legitimate initial consonant pair*/

{
	if ((b == 'r' || b == 'l') && (a == 'b' || a == 'c' || a == 'f'
		 || a == 'g' || a == 'k' || a == 'p' || a == 's' || a == 'v')) return true;
	else if (b == 'r' && (a == 'd' || a == 'm' || a == 't')) return true;
	else if ((b == 'm' || b == 'n') && (a == 'c' || a == 's')) return true;
	else if (a == 'j' && b == 'm') return true;
	else if (a == 'z' && (b == 'b' || b == 'v')) return true;
	else if ((a == 'c' || a == 's') && (b == 'k' || b == 'p' || b == 't')) return true;

	/*The additional pairs dj dz tc sv and ts are permitted*/

	else if ((a == 'd') && (b == 'j' || b == 'z')) return true;
	else if (a == 't' && (b == 'c' || b == 's')) return true;
	else if (a == 's' &&  b == 'v') return true;
	
	return false; /*All others are forbidden*/
}

const char * ComplexMaker::UnmangleForm( const char inForm[], string &out )
{
	assert(inForm != 0);
	
	const char *in = inForm;
	out.clear();
	bool looping = true;
	
	while (looping) {
		switch(*in) {
		
			// characters to ignore
		case '<':
		case '(':
		case ')':
		case '/':
			break;
			
			// characters to terminate on
		case '>':
		case '=':
		case '\0':
			looping = false;
			break;
			
			// characters to convert
		case '+':
		//	if ( (in[1] == 'y' || in[1] == 'r' || in[1] == 'n') && in[2] == '+' )
			if ( Letter::IsHyphen(in[1]) && in[2] == '+' )
				in++;	// it's a "hyphen": skip it
			else
				out.push_back(' ');	// it's a space
			break;
		
		default:
			// otherwise, copy
			out.push_back(*in);
			break;
		}
		in++;
	}
		
	return out.c_str();
}

#ifdef __MWERKS__
#pragma mark	=== ComplexMaker methods ===
#endif

ComplexMaker::ComplexMaker()
	: bestScore(0)
{
}

ComplexMaker::ComplexMaker(const char *words)
	: bestScore(0)
{
	SetWords(words);
}

ComplexMaker::ComplexMaker(const unsigned char *words)
	: bestScore(0)
{	// construct from a pascal string
	char buf[256];
	memmove(buf, words+1, words[0]);
	buf[words[0]] = '\0';
	
	SetWords(buf);
}

void ComplexMaker::Clear()
{
	// forget anything from the last use
	words.clear();
	affix.clear();
	bestWord.clear();
	bestForm.clear();
	bestScore = 0;
}

void ComplexMaker::SetWords( const char *wds )
{
	// read in the words from wds
	
	Clear();
	
	while (*wds != '\0') {
	
		// read in the next word
		Word w;
		wds = w.Set(wds);
		if (w.Get().size() > 0)
			words.push_back(w);
		
		// skip any non-word stuff
		while (*wds && !Letter::IsWordLetter(*wds)) wds++;
	}
	
	// set the affix data
	affix.resize(words.size());
	for (int i=0; i<words.size(); i++)
		affix[i].Set(words[i], i==(words.size()-1));
	
	if ( CanGetBest() )
		FindBest();
}

bool ComplexMaker::CanGetBest()
{
	// We can do our thing if:
	//		The CPrimSet has been properly initialized.
	//		There is more than one word in our list.
	return CPrimSet::Good() && words.size() > 1;
}

int ComplexMaker::GetComplexity()
{
	// the "complexity" is defined as the number of 
	// predicates that go into a complex.
	// This is used to make type entries for the LOD.
	int cnt = 0;
	for (int i=0; i<words.size(); i++)
		if (Word::IsPredicate(words[i].Get().c_str())) 
			cnt++;
	
	return cnt;
}

void ComplexMaker::FindBest()
{
	// Find the best complex from the current affixes.
	assert( CanGetBest() );
	
	// call MakeWord() on each permutation of affixes
	AffixPermutation p(affix.size());
	bestWord.clear();
	bestForm.clear();
	bestScore = 0;
	
	for (int i=0; i< p.size(); i++) p[i] = 0;	// initialize the permutation
	
	do {
		#if TRACE > 1
			cout << "Permutation:";
			for (int i=0; i< p.size(); i++)
				cout << ' ' 
					 << affix[i].Get(p[i]).Get() << '/' 
					 << TYPE(affix[i].Get(p[i]).GetType()) 
					 << flush;
		#endif
		
		MakeWord(p);
		
		#if TRACE > 1
			cout << endl;
		#endif
		
	} while (NextPermutation(p));
	
	#if TRACE > 1
		cout << endl;
	#endif
	#if TRACE > 0
		cout << " Best Word: " << bestWord 
			 << " Best Score: " << bestScore 
			 << " Best Form: " << bestForm << endl << endl;
	#endif
}

bool ComplexMaker::NextPermutation(AffixPermutation &p)
{
	// increment the permutation
	// return true until we're done
	int i = 0;
	while ( i < p.size() ) {
		if ( (++p[i]) >= affix[i].Count() ) {
			p[i] = 0;
			i++;
		} else
			break;
	}
	
	return i < p.size();
}

void ComplexMaker::MakeWord(const AffixPermutation &p)
{
	// make a trial word
	
	// penalties
	const long yOff = 60;  // A score penalty for a 'y hyphen'
	const long nrOff = 58; // The score penalty for an 'n' or 'r' hyphen
	const long cccOff = 14; // the score penalty for a three-consonant string.
	const long diCVvOff = 20; // The score penalty for two diCVvs in a row.
	
	string word, form;
	long score = 0;
	
	Affix::Type curType, prevType;
	curType = prevType = Affix::none;
	
	char c,d,e;
	c = d = e = '\0';
	
	const int first = 0;
	const int last = p.size()-1;
	
	#if TRACE > 2
		cout << endl;
		cout << " Subscores:";
	#endif
	
	// loop over the affixes
	for (int n = 0; n <= last; n++) {
		
		const Affix &aff = affix[n].Get(p[n]);	// the affix under consideration
		curType = aff.GetType();
		const string & text = aff.Get();
		int len = text.size();
		
		// Word never ends in a CVC. 
		// (We should never get a terminal CVC, but better safe than sorry.)
		if (n==last && curType == Affix::CVC) return;
		
		d = text[0];	// first letter of the affix
		e = text[1];	// second letter of the affix
		
		if (prevType != Affix::none) {
			// if there was a previous affix, test the affix joint
			
			/* If the new component is a borrowing, the 
				previous part complex must add 'y' unless it is 
				already 'y'.  Lettorals must add both 'r' and 'y' */
			if ( c != 'y' && aff.IsBorrowed() ) {
				if (prevType == Affix::letr) {
					word += 'r'; form += "+r";
				}
				word += 'y'; form += "+y";
				prevType = Affix::CCV; // Equivalent to CCV which can link anyhow.
			} else if (c != 'y' && prevType==Affix::letr) {
				// as affixes may also be letterals,
				// letterals must link with 'y' instead of 'r'
				word += 'y'; form += "+y";
			}
			
			switch(prevType) {
			case Affix::CCV:	// No action needed. CVV binds with anything
				break;
				
			case Affix::CVC:
			
			/* We check for an invalid cc junction, an invalid ccc, or if the cc
			junction would form an initial pair at the 0,1 junction of a >2 component complex
			(tosmabru/(pa)slinkui check).  All need hyphens */
			
				if (!IsValidJunction(c,d) || IsBadCccJoint(c,d,e) || (IsInitialPair(c,d)
					 && ( (n==1 && last>1) || (last==1 && len==5)))){
					word += 'y'; form += "+y";
					score -= yOff;
				} else if ( curType == Affix::CCV || curType == Affix::CCVCV || 
							curType == Affix::CVCCV ) {// CVCCV to match McIvor's code -- bug?
					score -= cccOff; // CCC penalty.
					#if TRACE > 3
						cout << " (cccOff)";
					#endif
				}
				break;
			
			case Affix::diCVV: 
				// Disyllabic CVVs must always have one vowel stressed
				// which means one V must be penultimate.
				if (n != last || curType != Affix::CCV || len != 3) {
					#if TRACE > 3
						cout << " (bad diCVV)";
					#endif
					return;
				}
				
			case Affix::diCVv: 
				// A string of disyllabic CVvs is penalized
				if (curType==Affix::diCVv) {
					score -= diCVvOff;
					#if TRACE > 3
						cout << " (diCVvOff)";
					#endif
				}
				// fall through...

			case Affix::monoCVV:
			case Affix::monoCVv:
				// Only CVVCCV can remain unhyphenated of words beginning with CVV
				if (n==1 && (n != last || !(curType == Affix::CCV && len == 3))){
					word += ((d == 'r') ? "n" : "r" );
					form += ((d == 'r') ? "+n" : "+r" );
					score -= nrOff;  // punish the hyphen
					#if TRACE > 3
						cout << " (nrOff)";
					#endif
				}
				break;
			default:
				break;
			}
		}	// end if prevType != none
		
		// append this affix to the trial
		word += text;
		if (n>first) form += '+';
		form += aff.GetForm();
		score += aff.GetScore();
		
		#if TRACE > 2
			cout << " " << score << flush;
		#endif
		
		// prepare for the next loop
		c = word[word.size()-1]; 	// last letter of previous loop
		prevType = curType;
	}	// end for(n=0; n<=last; n++)
	
	// remember the best one
	if (score > bestScore) {
		bestScore = score;
		bestWord = word;
		bestForm = form;
	}
	
	#if TRACE > 1
		cout << endl;
		cout << " Word: " << word << " Score: " << score << " Form: " << form;
	#endif
	
	// done with current permutation
}
