/*
	Word.cpp
	
	class Word
	
	Encapsulates routines that deal with a Loglan word.
	
	James Jennings * April 18, 2000
*/

#include "Word.h"
#include "Letter.h"
#include <string.h>

#ifdef __MWERKS__
#pragma mark	=== static utility functions ===
#endif

bool Word::IsName(const char *w)
{
	// A word is a name if it ends with a consonant.
	char c = w[ strlen(w) - 1 ];
	return Letter::IsConsonant(c);
}

bool Word::IsPredicate(const char *w)
{
	// A word is a predicate if it has a double consonant and ends in a vowel.
	bool lastWasC = false;
	bool hasCC = false;
	
	while (*w != '\0' && isalpha(*w)) {
		bool isC = Letter::IsConsonant(*w);
		if ( isC && lastWasC) hasCC = true;
		lastWasC = isC;
		w++;
	}
	return hasCC && Letter::IsVowel(*(w-1));
}

bool Word::IsPrimitive(const char *w)
{
	// Either CCVCV or CVCCV.
	// Note that these are "strict" tests. No "borrowed" letters or semi-vowels.
	
	if (strlen(w) != 5) return false;
	
	if ( ! Letter::IsConsonant(w[0]) ) return false;
	if ( ! Letter::IsConsonant(w[3]) ) return false;
	if ( ! Letter::IsVowel(w[4]) ) return false;
	
	if ( Letter::IsVowel(w[2]) && Letter::IsConsonant(w[1]) ) return true;	// CCVCV
	if ( Letter::IsVowel(w[1]) && Letter::IsConsonant(w[2]) ) return true;	// CVCCV
	
	return false;
}

bool Word::IsLetter(const char *w)
{
	/*
		Letter words are of the form:
		
		Consonants:
				Latin:		Greek:
		lower:	Cei			Ceo
		upper:	Cai			gaoCeo
		
		Vowels:
				Latin:		Greek:
		lower:	Vfi			Vzi
		upper:	Vma			gaoVfi
		
		The Cs and Vs are non-strict. (Include borrowed letters and semi-vowels.)
	
	*/
	const int kMaxLetterLength = 6;
	bool result = false;
	
	// copy to lower case
	char c[kMaxLetterLength+1];
	StringCopyToLowerCase( c, w, kMaxLetterLength+1 );
	
	if (Letter::IsConsonant(c[0], false)) {
		if (strncmp(c, "gao", 3) == 0)	// greek upper case
			result =	(Letter::IsConsonant(c[3], false) && strncmp(c+4, "eo", 2) ) ||
						(Letter::IsVowel(c[3], false) && strncmp(c+4, "fi", 2) );
		else
			result = (	strncmp( c+1, "ei", 2) == 0 ||	// latin lower case
						strncmp( c+1, "ai", 2) == 0 ||	// latin upper case
						strncmp( c+1, "eo", 2) == 0 );	// greek lower case
	} else if (Letter::IsVowel(c[0], false)) {
		result = (	strncmp( c+1, "fi", 2) == 0 ||	// latin lower case
					strncmp( c+1, "ma", 2) == 0 ||	// latin upper case
					strncmp( c+1, "zi", 2) == 0 );	// greek lower case
	}
	
	return result;
}

bool Word::IsBorrowing(const char *w)
{
	// A crude test but good enough for our purposes at the moment.
	// (Cribbed directly from RAM's makeword.c although I don't completely understand it.)
	short i;
	i = strlen(w);
	if (i==4 && !Letter::IsVowel(*w)) return false;
	if (i!=5 && i!=3) return true;
	if (Letter::IsVowel(*w) || Letter::IsVowel(w[3])) return true;
	return false;
}

const char * Word::PushUntilEndOfWord( string &dest, const char *w )
{
	// Allow only one word to be passed to dest.
	// A word may have a 'close-comma' embedded it in.
	if (w != 0)
		while(Letter::IsWordLetter(*w) || (*w == ',' && Letter::IsWordLetter(*(w+1))))
			dest.push_back( *(w++) );
	
	return w;
}


#ifdef __MWERKS__
#pragma mark	=== Word methods ===
#endif

Word::Word( const char *w )
{
	Set(w);
}
	
const char * Word::Set(const char *w) 
{
	// Set the word from the string.
	// Return a pointer to where we left off.
	word.clear();
	return PushUntilEndOfWord( word, w );
}

const string &	Word::Get() const 
{
	return word; 
}
