/*
	CEntryParser.cpp
	
	Class for parsing a dictionary entry so we may extract the meaningful pieces.
	
	by James Jennings
	Started March 28, 1999
*/

#pragma once

#include <assert.h>
#include "CEntryParser.h"

void CEntryParser::PutChar(char ch)
{
	switch((unsigned char)ch) {
	case '-':	// hyphens get turned into spaces
		mText += ' ';	// (should these be non-breaking spaces?)
		break;
	case '_':	// underscores get turned into hyphens
		mText += '-';
		break;
	case char_StylePhrase:
	case char_StyleGrammar:
	case char_StyleEmphasis:
	case char_EndStyleEmphasis:
		// ignore imbedded style commands
		break;
	default:
		mText += ch;
		break;
	}
}

void CEntryParser::PutString(char *s)
{
	assert(s!=0);
	while (*s != '\0')
		PutChar(*(s++));
}

void CEntryParser::PutNewLine()
{
	PutChar('\n');	// Do I need to handle cross platform issues better?
}

char * CEntryParser::GetPart(EParsePart inPart, long inIndex, char *ioBuffer, long inBufferLength)
{	// get the text that belongs to a part
	
	assert((0<=inIndex && inIndex < definition.size()) || inIndex == 0);
	
	Range *r = GetRange(inPart, inIndex);
	
	if (r != 0 && (r->size() > 0) ) {
		long len = r->size();
		if (len >= inBufferLength)
			len = inBufferLength-1;
		memcpy( ioBuffer, mText.c_str() + r->start, len);
		ioBuffer[len] = '\0';
	} else {
		strncpy(ioBuffer, "<no data>", inBufferLength);
	}
	
	return ioBuffer;
}

unsigned char *	CEntryParser::GetPart(EParsePart inPart, long inIndex, unsigned char * ioPString)
{	// get a Pascal string version of the text that belongs to the part

	assert((0<=inIndex && inIndex < definition.size()) || inIndex == 0);
	
	Range *r = GetRange(inPart, inIndex);
	
	if (r != 0 && (r->size() > 0) ) {
		long len = r->size();
		if (len >= 255)
			len = 255;
		ioPString[0] = len;
		memcpy( ioPString+1, mText.c_str() + r->start, len);
	} else {
		ioPString[0] = 0;	// empty pascal string
	}
	
	return ioPString;
}

CEntryParser::Range * CEntryParser::GetRange(EParsePart inPart, long index)
{	// helper function for GetPart()
	
	assert((0<=index && index < definition.size()) || index == 0);
	
	Range *r = 0;
	
	switch(inPart) {
	case part_Header:		r = &header; 		break;
	case part_Word:			r = &word; 			break;
	case part_WordInfo:		r = &info;			break;
		case part_Affixes:	r = &affixes;		break;
		case part_Etymology:r = &etymology;		break;
	case part_TypeField:	r = &typeField;		break;
		case part_Percentage:r = &percentage;	break;
		case part_WordType:	r = &type;			break;
	case part_Attribution:	r = &attribution;	break;
		case part_Author:	r = &author;		break;
		case part_Date:		r = &date;			break;
		case part_Rank:		r = &rank;			break;
	case part_Definition:	r = &definition[index].all;		break;
		case part_Phrase:	r = &definition[index].phrase;	break;
		case part_Grammar:	r = &definition[index].grammar;	break;
		case part_DefinitionText:r = &definition[index].text;break;
		case part_CaseTags:	r = &definition[index].tags;	break;
	case part_Footer:		r = &footer;		break;
	// add more parts as needed
	default:	r = 0;		break;	// no such part
	}
	
	return r;
}

// === Recursive Descent Parse ===

void CEntryParser::ThrowOnUnexpectedEnd(int ch)
{
	if (ch == char_EndField) throw EndOfFieldException();
	if (ch == char_EndDefinition) throw EndOfDefinitionException();
	if (IsEndOfEntry(ch)) throw EndOfEntryException();
}

bool CEntryParser::IsEndOfEntry(int ch)
{
	return	ch == 0x0A;	// entry separator
}

void CEntryParser::Parse(istream &in)
{
	// Start of a recursive descent parse.
	// Rule:	entry := word ',' [ header ] definition [ definition ]* [ footer ]
	
	try {
		// get word, keep trailing comma
		ParseChunk(in, word, char_EndWord, true, false);
		
		bool usingLE = ParseHeader(in);
		
		Definition def;
		while (ParseDefinition(in, def, usingLE))
			definition.push_back(def);
		
		ParseFooter(in);	// Used in: ...
		
	}
	catch (EndOfEntryException ex) {}
	catch (EndOfDefinitionException ex) {
		assert(false /*unexpected end of definition*/);
	}
	catch (EndOfFieldException ex) {
		assert(false /*unexpected end of field*/);
	}
}

char CEntryParser::ParseChunk(istream &in, Range &chunk, 
		const unsigned char stopAt, bool add, bool include)
{
	// general purpose chunk parsing function
	// It scans/copies text until:
	//		it finds the stopAt character
	//		it finds the start of a new definition
	//		it finds the end of the entry
	// When it stops it can either:
	//		add the stop character to the text
	//			include it in the chunk
	//			don't include it
	//		don't add it to the text
	//			ignore it completely
	//			unget it (include in the next chunk)
	
	// remember the starting position
	unsigned char ch;
	chunk.start = chunk.end = mText.size();
	
	while ( bool(in.get((char&)ch) ) ) {
		if (ch == stopAt) {
			if (add) {		// add the delimiter in the text
				PutChar(ch);
				if (include)	// include the delimiter in the chunk
					chunk.end = mText.size();
			} else {		// don't add
				if (include)
					in.unget();	// include in the next chunk
			}
			break;
		}
		ThrowOnUnexpectedEnd(ch);
		
		PutChar(ch);
		chunk.end = mText.size();
	}
	return 0;	// zero means we stopped at the expected delimiter
}

bool CEntryParser::ParseHeader(istream &in)
{
	// Parse an entry header. Return true if a header was found.
	// Rule:	header := [wordInfo '|' <unused> '|' wordType '|' attribution] ''
	//	Note: The optional part is only in the LE dictionary.
	
	// make sure we have a header (only the LE dictionary entries have one)
	unsigned char ch = in.peek();
	if (ch == char_EndHeader) {
		in.ignore(1);
		PutNewLine();
		return false;
	}
	assert(ch == char_StyleEmphasis);
	
	header.start = header.end = mText.size();
	
	try {
	
		// get wordInfo
		ParseChunk(in, info, char_EndField, false, false);
		if (info.size()>0) {
			PutChar(' ');
			// split wordInfo into affixes and etymology
			size_type where = mText.find(char_StartEtymology, info.start);
			if (info.contains(where)) {
				affixes.set(info.start, where-1);
				etymology.set(where, info.end);
			} else {
				affixes = info;
			}
		}
		
		// skip unused field
		Range dummy;
		ParseChunk(in, dummy, char_EndField, false, false);
		
		// get word type
		ParseChunk(in, typeField, char_EndField, false, false);
		if (typeField.size()>0) {
			PutChar(' ');
			// split word type into %age and type
			size_type where = mText.find('%', typeField.start);
			if (typeField.contains(where)) {
				percentage.set(typeField.start, where+1);
				type.set(where+2, typeField.end);
			} else {
				type = typeField;
			}
		}
		
		// get attribution
		ParseChunk(in, attribution, char_EndHeader, false, false);
		if (attribution.size()>0) {
			PutChar(' ');
			// split attribution into author, date, and rank at the spaces
			size_type where = mText.find(' ', attribution.start);
			if (attribution.contains(where)) {
				author.set(attribution.start, where);
				date.set(where+1, attribution.end);
				
				size_type where = mText.find(' ', date.start);
				if (date.contains(where)) {
					date.set(date.start, where);
					rank.set(where+1, attribution.end);
			}
				
			}
		}
	}
	catch (EndOfEntryException ex) {
		header.end = mText.size();
		throw;
	}
	catch (EndOfDefinitionException ex) {}
//	catch (EndOfFieldException ex) {}
	header.end = mText.size();
	
	PutNewLine();
	
	return true;
}

bool CEntryParser::ParseDefinition(istream &in, Definition &def, bool usingLE)
{
	// Parse a entry definition. Return true if there was something to parse.
	// LE Rule:	definition := phrase '' grammar '|' text '|' tags '' 
	// EL Rule:	definition := phrase '' grammar ' ' text '' 
	
	unsigned char ch;
	// make sure we have a definition
	ch = in.peek();
	if (ch != char_StylePhrase &&	// if there's a phrase, we'll see the style mark
		ch != char_EndPhrase) 		// or, we'll see the delimiter
			return false;			// otherwise, we don't have a definition
	
	PutString("   ");	// indent
	
	def.all.start = def.all.end = mText.size();
	
	try {
		// get phrase
		if (usingLE)
			ParseChunk(in, def.phrase, char_EndPhrase, false, false);
		else
			ParseChunk(in, def.phrase, char_StyleGrammar, false, false);
		if (def.phrase.size()>0) PutChar(' ');
		// get grammar
		if (usingLE)
			ParseChunk(in, def.grammar, char_EndField, false, false);
		else
			ParseChunk(in, def.grammar, ' ', false, false);
		if (def.grammar.size()>0) PutChar(' ');
		// get text
		if (usingLE)
			ParseChunk(in, def.text, char_EndField, false, false);
		else
			ParseChunk(in, def.text, char_EndDefinition, false, false);
		if (def.text.size()>0) PutChar(' ');
		// get tags
		if (usingLE) {
			ParseChunk(in, def.tags, char_EndDefinition, false, false);
			if (def.tags.size()>0) PutChar(' ');
		}
	}
	catch (EndOfEntryException ex) {
		def.all.end = mText.size();
		throw;
	}
	catch (EndOfDefinitionException ex) {}
	
	def.all.end = mText.size();
	
	PutNewLine();
	
	return def.all.size() > 0;
}

void CEntryParser::ParseFooter(istream &in)
{
	// Parse the footer, if any.
	// Rule:	footer := "Used in:" word [ ';' word ]* ''
	try {
		ParseChunk(in, footer, char_EndFooter, false, false);
	}
	catch (EndOfDefinitionException ex) {}
}
