/*
	CEntryParser.h
	
	Class for parsing a dictionary entry so we may extract the meaningful pieces.
	
	by James Jennings
	Started March 28, 1999
*/

#pragma once

#include <vector>
#include <string>
#include <stdio.h>
#include <istream>

using namespace std;

/*
		List of characters that have special meanings.
*/
enum {
//	char_WordEnd			= ',',
//	char_HeaderDelimiter	= '|',
	char_StartEtymology		= '<',
	char_EndEtymology		= '>',
	char_AttributionDelimeter = ' ',
//	char_StartPhrase		= 0xB2, 	// '' Mac less than or equals
//	char_StartGrammar		= 0xB3,		// '' Mac greater than or equals
//	char_DefinitionDelimeter= '|',
	char_StartTag			= '[',
	char_EndTag				= ']',
	
	char_EndEntry			= 0x0A,		//	LF
	char_EndEntry2			= 0x0D,		//	CR	not actually used
	char_EndHeader			= 0xB2, 	// '' Mac less than or equals,
	char_EndDefinition		= 0xB2, 	// '' Mac less than or equals,
	char_EndFooter			= 0xB2, 	// '' Mac less than or equals,
	char_EndField			= '|',
	char_EndWord			= ',',
	char_EndPhrase			= 0xB3,		// '' Mac greater than or equals
	
	char_StylePhrase		= 0xC7,		// '' Mac left French quote
	char_StyleGrammar		= 0xC8,		// '' Mac right French quote
	char_StyleEmphasis		= 0xD6,		// '' Mac divide
	char_EndStyleEmphasis	= 0xC0		// '' Mac Spanish left question mark
	};


/*
	EParseParts
		Codes for the meaningful parts of an entry
*/
typedef enum {
		// 1st level parse
		part_Word,		// the word the entry is for
		part_Header,	// meta information about the word (LE dictionary only)
		part_Definition,// a definition (may be more than one)
		part_Footer,
		// 2nd level parse
		part_FirstHeaderPart,
			part_WordInfo = part_FirstHeaderPart,		// affixes, etymology
			part_Unused,
			part_TypeField,		// Eg: "C-Prim"
			part_Attribution,	// Author, Date, Rank
		part_NoMoreHeaderParts,
		part_FirstDefinitionPart,
			part_Phrase,		// eg: 'lopo --'
			part_Grammar,		// eg: '(3a)'
			part_DefinitionText,
			part_CaseTags,		// eg: '[B-D]' (LE dictionary only)
		part_NoMoreDefinitionParts,
		// 3rd level parse (header only)
		part_FirstWordInfoPart,
			part_Affixes,
			part_Etymology,
		part_FirstTypeFieldPart,
			part_Percentage,
			part_WordType,
		part_FirstAttributionPart,
			part_Author,		// eg: 'L4' or 'JCB'
			part_Date,			// eg: "'77"
			part_Rank,			// Eaton rank, eg: '2.4'
		//
		part_NoMoreParts
		} EParsePart;

class CEntryParser {
protected:
	typedef string::size_type size_type;
	class Range {
	public:
		Range() : start(0), end(0) {}
		void
		set(size_type s, size_type e) 
			{ start = s; end = ((s<e) ? e : s); }
		size_type
		size() { return end-start; }
		bool
		contains(size_type s) 
			{ return s != string::npos && start <= s && s <= end; }
		size_type	start,end;
	};
	class Definition {
	public:
		Range	all,
				phrase,
				grammar,
				text,
				tags;
	};

	// exceptions thrown entirely inside the parse
	// Note that the EndOfFieldException is used for the end of the header as well 
	class EndOfFieldException {};
	class EndOfDefinitionException : public EndOfFieldException {};
	class EndOfEntryException : public EndOfDefinitionException {};
	
public:
	//			CEntryParser() {}
	
		// formatting
	virtual void	PutChar(char ch);
	virtual void	PutString(char *s);
	virtual void	PutNewLine();
		// accessing
	virtual int		CountDefinitions() { return definition.size(); };
	virtual char *	GetPart(EParsePart inPart, long index, 
						char *ioBuffer, long inBufferLength);	// as C string
	virtual unsigned char *	GetPart(EParsePart inPart, long inIndex, 
						unsigned char * inPString);				// as Pascal string
		// parsing
	virtual void	Parse(istream &in);
	
protected:
	virtual bool	IsEndOfEntry(int ch);
	virtual void	ThrowOnUnexpectedEnd(int ch);
	virtual	char	ParseChunk(istream &in, Range &chunk, 
						const unsigned char stopAt, bool pass, bool include);

	virtual bool	ParseHeader(istream &in);
	virtual bool	ParseDefinition(istream &in, Definition &def, bool usingLE);
	virtual void	ParseFooter(istream &in);
	
	virtual Range * GetRange(EParsePart inPart, long index);
	
	string mText;	// the text after parsing
	
	// the parts of the parsed text
	Range	header,
				word,
				info,
					affixes,
					etymology,
				typeField,
					percentage,
					type,
				attribution,
					author,
					date,
					rank;
	vector<Definition> definition;
	Range	footer;
};
