// ===========================================================================
//	 CWordMaker.cp		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================

/*
	JPJ:
	Neatened things up a little:
		Made functions static and const as appropriate
		Turned all Booleans into bool (cross platform standard C++)
		
*/

#include <string.h>

#include "CWordMaker.h"
#include "CPrimSet.h"
#include "CWord.h"
#include "OldUtility.h"

/*
	Choose the TRACE level for debugging FindBest()
	0: Trace is off
	1: Show best score
	2: Trace each permutation
	3: Trace each affix
	4: Trace each decision.
*/
#define TRACE	0

#if TRACE > 0
	#include <iostream>
	const char *affixTypeNames[] = {
			"none",
			"CCV",
			"CVC",
			"diCVv",	// di-syllabic non-equal vowels
			"diCVV",	// di-syllabic equal vowels
			"monoCVv",	// mono-syllabic non-equal vowels
			"monoCVV",	// mono-syllabic equal vowels
			"letr",		// lettoral
			"dummy",
			"CCVCV",	// primitive
			"CVCCV",	// primitive
			"other"
			};
	static const char *TYPE( CAffix::AffixType t ) { return affixTypeNames[t+1]; }
#endif

// ===========================================================================
//	 Constructors		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================

CWordMaker::CWordMaker()	// Default constructor

{	Init();
	}
	
CWordMaker::CWordMaker(const char *words)

{	Init();
	SetWords(words);
	}
	
CWordMaker::CWordMaker(const unsigned char *words)

{	// construct from a pascal string
	char buf[256];
	memmove(buf, words+1, words[0]);
	buf[words[0]] = '\0';
	
	Init();
	SetWords(buf);
	}
	
// ===========================================================================
//	 Destructor		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================

//CWordMaker:: ~CWordMaker() {}
	
// ===========================================================================
//	 Init				1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================

void CWordMaker::Init()

{
	bestScore = testScore = 0;
	primCount = 0;
	bestWord[0] = testWord[0] = '\0';  //Set preliminary values
	}
	
// ===========================================================================
//	 SetWords			1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Separate out the individual primitives in a String

void CWordMaker::SetWords(const char *wordList) 
{
	assert(wordList != 0);
	const int kBufferLength = 256;
	char s[kBufferLength];
	
	// ParseWordList() can change it's input, so we make a copy of it,
	// converting to lower case as we go.
	StringCopyToLowerCase( s, wordList, kBufferLength );
	
	// crunch on the input
	ParseWordList(s);
	if (primCount > 1 && CPrimSet::Good()) {
		// if we have more than one primitive AND the primitive data is available,
		// combine the primitives
		CalcAffixes();
		FindBest();
	} else if (primCount >= 1) {
		// if there is only one primitive, or primitive data wasn't available,
		// use the input straight
		strncpy(bestWord, s, kWordLength);
	} else {
		// there were no primitives: probably a user error
		bestWord[0] = '\0';	// empty string
	}
}

// ===========================================================================
//	 ParseWordList		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Separate out the individual primitives in a String

void CWordMaker::ParseWordList(char *theString) 
{
	/*
		Loop through the string, extracting words with strtok().
		Assign each word to one of our CWord objects.
		Notes:
			CWord::SetString() checks word length and does prim.set lookup.
			strtok() changes it's input string.
	*/
	short i = 0;
	for ( char *s = strtok(theString," ,"); s != 0; s = strtok(NULL,", ")) {
		wd[i++].SetString(s);		// set value in the appropriate CWord instance
		if (i>=kMaxWords) break;	// Our array is full. Ignore the rest. -- JPJ
	}
	
	primCount = i;  // This is the total number of primitives to include in the complex
		
	}

// ===========================================================================
//	 GetAffixes		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Obtain the affixes for each of the primitives in the set

void CWordMaker::CalcAffixes()

{	short i;	
	for (i = 0; i< primCount;++i) {
		wd[i].CalcAffixes(primCount-1 == i);
		}
	}

	
// ===========================================================================
//	 FindBest			1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Find the best complex using one affix from each primitive

char * CWordMaker::FindBest()

{
	short i,j,k,l,m,n;
	short index[5];
	bestWord[0] = '\0';		// Initialize the values
	bestScore = testScore = 0;
	n = primCount;
	for (i=0; i<wd[0].GetAffCount();++i) {
		index[0] = i;
		for (j=0; j<wd[1].GetAffCount(); ++j) {
			index[1] = j;
			if (n==2) MakeWord(index);  // If only two primitives in set, make the word now
			for (k = 0; k<wd[2].GetAffCount(); ++k) {
				index[2] = k;
				if (n==3) MakeWord(index);  // likewise if three primitives
				for (l = 0; l<wd[3].GetAffCount(); ++l) {
					index[3] = l;
					if (n==4) MakeWord(index); // or if four
					for (m = 0; m<wd[4].GetAffCount(); ++m) {
						index[4] = m;
						MakeWord(index);	// finally, if five.						
						}
					}
				}
			}
		}
	#if TRACE > 1
		cout << endl;
	#endif
	#if TRACE > 0
		cout << " Best Word: " << bestWord 
			 << " Best Score: " << bestScore 
			 << " Best Form: " << bestForm << endl << endl;
	#endif
	return bestWord;			
	}

// ===========================================================================
//	 MakeWord			1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Form a word from affixes.  Index is a set of up to 5 affixes from which to
// make a test word.

void CWordMaker::MakeWord (short *theIndex)

{	assert(theIndex != 0);
	
	CAffix myAff;
	const short yOff = 60;  // The score penalty for a 'y hyphen'
	const short nrOff = 58; // The score penalty for an 'n' or 'r' hyphen
	const short cccOff = 14; // The score penalty for a three-consonant string.
	const short cvvOff = 20; // The score penalty for two diCVvs in a row.
	
	#if TRACE > 1
		cout << "Permutation:";
		for (int i=0; i< primCount; i++)
			cout << ' ' 
				 << wd[i].GetPrimAffix(theIndex[i]).GetAffixText() << '/' 
				 << TYPE(wd[i].GetPrimAffix(theIndex[i]).GetAffixType()) 
				 << flush;
		cout << endl;
	#endif

	CAffix::AffixType curType, prevType;
	short n,last,len;
	char c,d,e;
	const char *aff;
	testWord[0] = '\0';	// Blank out word
	testForm[0] = '\0';
	testScore = 0;		// and its score
	prevType = CAffix::CCV;		// previous affix type
	last = primCount-1;	// get index of last component
	
	#if TRACE > 2
		cout << endl;
		cout << " Subscores:";
	#endif

	for (n=0; n<primCount; ++n) {	//for each component.
		myAff= wd[n].GetPrimAffix(theIndex[n]);	//for one affix value of each component
		curType = myAff.GetAffixType();		// get its type
		if (n==last && curType==CAffix::CVC) return;		// Complex cannot end with a consonant.
		aff = myAff.GetAffixText(); // text of the affix
		len = strlen(aff);
		c = testWord[strlen(testWord)-1];  // Last letter of current build.
		d = aff[0];	// First letter of affix to add
		e = aff[1];	// Second letter of affix to add.
		
/* If the new component is a borrowing, the previous part complex must add 'y' unless 
it is already 'y'.  Lettorals must add both 'r' and 'y' */

		if (c!='y' && n && IsBorrowed(aff)) {
			if (prevType==CAffix::letr) { 
				strcat(testWord,"r"); 
				strcat(testForm,"+r"); 
			}
			strcat(testWord,"y"); 
			strcat(testForm,"+y");
			prevType = CAffix::CCV; // Equivalent to CCV which can link anyhow.
			}
	// as affixes may also be letterals, letterals must link with 'y' instead of 'r'
		else if (c != 'y' && n && prevType==CAffix::letr) {
			strcat(testWord, "y"); // Letterals use 'y'
			strcat(testForm,"+y");
		}
		switch (prevType) {
		case CAffix::CCV:
			break; // No action needed
		case CAffix::CVC:
		
		/* We check for an invalid cc junction, an invalid ccc, or if the cc
		junction would form an initial pair at the 0,1 junction of a >2 component complex
		(tosmabru/(pa)slinkui check).  All need hyphens */
		
			if (!IsValidJunction(c,d) || IsBadCccJoint(c,d,e) || (IsInitialPair(c,d)
				 && ( (n==1 &&last>1) || (last==1 && len==5)))){
				strcat(testWord,"y");
				strcat(testForm,"+y");
				testScore -= yOff;
				}
			else if (curType== CAffix::CCV) testScore -=cccOff; // CCC penalty.
			break;
			
		// Like disyllabic CVVs must always have one vowel stressed
		case CAffix::diCVV: if (n != last || curType != CAffix::CCV || len != 3) return;
		
		// A string of disyllabic CVVs is penalized
		case CAffix::diCVv: if (curType==CAffix::diCVv) testScore -= cvvOff;
		
		// Only CVVCCV can remain unhyphenated of words beginning with CVV
		case CAffix::monoCVV:
		case CAffix::monoCVv:
			if (n==1 && (n != last || !(curType == CAffix::CCV && len == 3))){
				strcat (testWord,(d=='r')?"n":"r");
				strcat (testForm,(d=='r')?"+n":"+r");
				testScore -= nrOff;  // punish the hyphen
				}
			}
		strcat(testWord,myAff.GetAffixText()); // join them up
		if (n>0) strcat(testForm,"+");
		strcat(testForm,myAff.GetForm());
		testScore += myAff.GetAffixScore(); // add the score
		#if TRACE > 2
			cout << " " << testScore << flush;
		#endif
		prevType = curType;
		}
	if (testScore>bestScore) {  //Check if a new best word has been found.
	//	for (n=0; n<primCount;++n)  wd[n].PutBestAffix(theIndex[n]);
		bestScore =  testScore;
		strcpy (bestWord, testWord);
		strcpy (bestForm, testForm);
		}
	#if TRACE > 2
		cout << endl;
	#endif
	#if TRACE > 1
		cout << " Word: " << testWord << " Score: " << testScore << " Form: " << testForm;
		cout << endl;
	#endif
	}

// === static methods ===

const char *	CWordMaker::UnmangleForm( const char inForm[], char outString[])
{
	assert(inForm != 0 && outString != 0);
	
	const char *in = inForm;
	char *out = outString;
	bool looping = true;
	
	while (looping) {
		switch(*in) {
		
			// characters to ignore
		case '<':
		case '(':
		case ')':
		case '/':
			break;
			
			// characters to terminate on
		case '>':
		case '=':
		case '\0':
			looping = false;
			break;
			
			// characters to convert
		case '+':
			if ( (in[1] == 'y' || in[1] == 'r' || in[1] == 'n') && in[2] == '+' )
				in++;	// it's a "hyphen": skip it
			else
				*out++ = ' ';	// it's a space
			break;
		
		default:
			// otherwise, copy
			*out++ = *in;
			break;
		}
		in++;
	}
	
	*out = '\0';	// terminate the string
	
	return outString;
}

int CWordMaker::GetComplexity( const char inForm[] )
{
	// count the number of predicates in the input string
	const char *s = inForm;
	int cnt = 0;
	while(*s != 0) {
		if (CWord::IsPredicate(s)) cnt++;
		while(*s && isalpha(*s)) s++;	// skip to end of word
		while(*s && !isalpha(*s)) s++;	// skip delimeters
	}
	
	return cnt;
}

int CWordMaker::GetComplexity()
{
	// count the number of true predicates in the make word
	int i, cnt;
	for (i=0, cnt = 0; i<primCount; i++)
		if (wd[i].IsPredicate()) cnt++;
	
	return cnt;
}

// ===========================================================================
//	 IsBorrowed		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Check for a borrowing mainly by eliminating non borrowed word forms

bool CWordMaker::IsBorrowed(const char *theString)

{	short i;
	i = strlen(theString);
	if (i==4 && !IsVowel(*theString)) return false; // 4-letter borrowings begin with V
	if (i!=5 && i!=3) return true; // If not length 3 or 5
	
	// 5-letter primitives begin with a consonant CvCCv or CCvCv and have one at posn 4
	// JPJ: shouldn't we test for i==5?
	if (IsVowel(*theString) || IsVowel(theString[3])) return true; 
	return false;
	}

	
// ===========================================================================
//	 IsBadCccJoint		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// A certain number of CCCs are nonpronounceable. This checks for those.

bool CWordMaker::IsBadCccJoint(char a, char b, char c)

/*This routine examines two words a and b to see if the last consonant of word a,
 and the first two consonants of word b are in a list of forbidden ccc's */

{	char tst[4];
	const char *ccc[]={	// currently 20 CCCs
		"cdz","cvl","jdj", "jtc","jts","jvr","ndj","ndz","dcm","dct",
		"dts","kdz","pdz","svl","tvl","gts", "gzb","mzb","vts","njm",
		0};	// zero marks the end of the array : JPJ
	tst[0]=a;
	tst[1]=b;
	tst[2]=c;
	tst[3]='\0';
	for (int i = 0; ccc[i]!=0; ++i)
		if (strcmp(tst,ccc[i])==0) return(true); 
	return(false);	/* ccc string good */
	}


// ===========================================================================
//	 IsValidJunction	1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// A certain number of CCs are not accurately pronounceable. This checks for those.

bool CWordMaker::IsValidJunction(char a, char b)

/* Check if the joint between the last character of an affix a and the first character
 of the next affix b is legal*/

{	
/*the two characters cannot be identical, and the first cannot be 'h'*/

	if (a=='h'||a==b) return false;
	
/*Both characters cannot be spirants*/

	if (IsSpirant(a)&&IsSpirant(b)) return false;
	
/*If the first character is non-voiced, and the second is a voiced IsSpirant, the joint is illegal*/

	if (IsUnvoiced(a)&&(b=='j'||b=='z')) return false;
	
/*An unvoiced consonant cannot be followed by the equivalent voiced cons.*/

	if (a=='p' && b=='b') return false;
	if (a=='t'&&b=='d') return false;	
	if (a=='k'&&b=='g') return false;
	if (a=='f'&&b=='v') return false;
	if (a=='b'&&b=='j') return false;
	if (a=='s'&&b=='b') return false;

	return true; /*all others are valid*/
	}

	
// ===========================================================================
//	 IsInitialPair		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// An initial pair at a junction may have to be hyphenated. Check for these

bool  CWordMaker::IsInitialPair(char a, char b)

/*Check if the consonant pair ab is a legitimate initial consonant pair*/

{	if ((b=='r'||b=='l')&&(a=='b'||a=='c'||a=='f'
		||a=='g'||a=='k'||a=='p'||a=='s'||a=='v')) return true;
	else if (b=='r'&&(a=='d'||a=='m'||a=='t')) return true;
	else	if ((b=='m'||b=='n')&&(a=='c'||a=='s')) return true;
	else if (a=='j'&&b=='m') return true;
	else if (a=='z'&&(b=='b'||b=='v')) return true;
	else if ((a=='c'||a=='s')&&(b=='k'||b=='p'||b=='t')) return true;

	/*The additional pairs dj dz tc sv and ts are permitted*/

	else if ((a=='d')&&(b=='j'||b=='z')) return true;
	else if (a=='t'&&(b=='c'||b=='s')) return true;
	else if (a=='s' && b=='v') return true;
	else return false; /*All others are forbidden*/
	}
	

// ===========================================================================
//	 IsSpirant			1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// Junctions of spirants are forbidden.  Check if a consonant is one of these.

bool CWordMaker::IsSpirant(char a)

/* Check if character a is a IsSpirant*/

{	return (a=='c'||a=='j'||a=='s'||a=='z');
	}

// ===========================================================================
//	 IsUnvoiced		1994-1998 Loglan Institute Inc. All rights reserved.
// ===========================================================================
// 

bool CWordMaker::IsUnvoiced(char a)

/* Check for an unvoiced consonant */

{	return (a=='p'||a=='t'||a=='k'||a=='f'||a=='s'||a=='c');
	}


